"""privacy_preference_v2

Revision ID: f9b28f36b53e
Revises: f396c1f84b0f
Create Date: 2023-12-13 17:14:00.677493

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

from fides.api.models.privacy_notice import ConsentMechanism

revision = "f9b28f36b53e"
down_revision = "f396c1f84b0f"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "currentprivacypreferencev2",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "email",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "fides_user_device",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "phone_number",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("hashed_email", sa.String(), nullable=True),
        sa.Column("hashed_fides_user_device", sa.String(), nullable=True),
        sa.Column("hashed_phone_number", sa.String(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "preferences", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column("fides_string", sa.String(), nullable=True),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("email", name="last_saved_for_email"),
        sa.UniqueConstraint(
            "fides_user_device", name="last_saved_for_fides_user_device"
        ),
        sa.UniqueConstraint("phone_number", name="last_saved_for_phone_number"),
    )
    op.create_index(
        op.f("ix_currentprivacypreferencev2_created_at"),
        "currentprivacypreferencev2",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreferencev2_hashed_email"),
        "currentprivacypreferencev2",
        ["hashed_email"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreferencev2_hashed_fides_user_device"),
        "currentprivacypreferencev2",
        ["hashed_fides_user_device"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreferencev2_hashed_phone_number"),
        "currentprivacypreferencev2",
        ["hashed_phone_number"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreferencev2_id"),
        "currentprivacypreferencev2",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreferencev2_updated_at"),
        "currentprivacypreferencev2",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "lastservednoticev2",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "email",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "fides_user_device",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "phone_number",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("hashed_email", sa.String(), nullable=True),
        sa.Column("hashed_fides_user_device", sa.String(), nullable=True),
        sa.Column("hashed_phone_number", sa.String(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("served", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("email", name="last_served_for_email"),
        sa.UniqueConstraint(
            "fides_user_device", name="last_served_for_fides_user_device"
        ),
        sa.UniqueConstraint("phone_number", name="last_served_for_phone_number"),
    )
    op.create_index(
        op.f("ix_lastservednoticev2_created_at"),
        "lastservednoticev2",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednoticev2_hashed_email"),
        "lastservednoticev2",
        ["hashed_email"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednoticev2_hashed_fides_user_device"),
        "lastservednoticev2",
        ["hashed_fides_user_device"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednoticev2_hashed_phone_number"),
        "lastservednoticev2",
        ["hashed_phone_number"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednoticev2_id"), "lastservednoticev2", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_lastservednoticev2_updated_at"),
        "lastservednoticev2",
        ["updated_at"],
        unique=False,
    )
    op.drop_index(
        "ix_currentprivacypreference_privacy_preference_history_id",
        table_name="currentprivacypreference",
    )
    op.drop_constraint(
        "currentprivacypreference_privacy_preference_history_id_fkey",
        "currentprivacypreference",
        type_="foreignkey",
    )
    # Getting rid of this link - currentprivacypreference is deprecated and we need to drop this existing connection
    op.drop_column("currentprivacypreference", "privacy_preference_history_id")
    op.drop_index(
        "ix_lastservednotice_served_notice_history_id", table_name="lastservednotice"
    )
    op.drop_constraint(
        "lastservednotice_served_notice_history_id_fkey",
        "lastservednotice",
        type_="foreignkey",
    )
    # Getting rid of this link - lastservednotice is deprecated and we need to drop this existing connection
    op.drop_column("lastservednotice", "served_notice_history_id")
    op.add_column(
        "privacypreferencehistory", sa.Column("notice_key", sa.String(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("notice_mechanism", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("notice_name", sa.String(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("fides_string", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "tcf_preferences", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )
    op.alter_column(
        "privacypreferencehistory",
        "preference",
        type_=sa.String(),
        existing_type=sa.Enum(ConsentMechanism),
    )
    op.drop_index(
        "ix_privacypreferencehistory_feature", table_name="privacypreferencehistory"
    )
    op.drop_index(
        "ix_privacypreferencehistory_fides_user_device_provided__9961",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_provided_identity_id",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_purpose_consent",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_purpose_legitimate_interests",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_special_feature",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_special_purpose",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_system_consent",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_system_legitimate_interests",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_vendor_consent",
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        "ix_privacypreferencehistory_vendor_legitimate_interests",
        table_name="privacypreferencehistory",
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_notice_key"),
        "privacypreferencehistory",
        ["notice_key"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_notice_mechanism"),
        "privacypreferencehistory",
        ["notice_mechanism"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_notice_name"),
        "privacypreferencehistory",
        ["notice_name"],
        unique=False,
    )
    op.drop_constraint(
        "privacypreferencehistory_provided_identity_id_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_constraint(
        "privacypreferencehistory_fides_user_device_provided_identi_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_constraint(
        "privacy_preference_served_notice_fk",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_column("privacypreferencehistory", "vendor_consent")
    op.drop_column("privacypreferencehistory", "purpose_consent")
    op.drop_column("privacypreferencehistory", "purpose_legitimate_interests")
    op.drop_column("privacypreferencehistory", "system_legitimate_interests")
    op.drop_column("privacypreferencehistory", "fides_user_device_provided_identity_id")
    op.drop_column("privacypreferencehistory", "special_feature")
    op.drop_column("privacypreferencehistory", "special_purpose")
    op.drop_column("privacypreferencehistory", "tcf_version")
    op.drop_column("privacypreferencehistory", "system_consent")
    op.drop_column("privacypreferencehistory", "relevant_systems")
    op.drop_column("privacypreferencehistory", "vendor_legitimate_interests")
    op.drop_column("privacypreferencehistory", "feature")
    op.drop_column("privacypreferencehistory", "provided_identity_id")
    op.add_column(
        "servednoticehistory", sa.Column("notice_key", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("notice_mechanism", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("notice_name", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("served_notice_history_id", sa.String(), nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("tcf_served", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    )
    op.drop_index("ix_servednoticehistory_feature", table_name="servednoticehistory")
    op.drop_index(
        "ix_servednoticehistory_fides_user_device_provided_identity_id",
        table_name="servednoticehistory",
    )
    op.drop_index(
        "ix_servednoticehistory_provided_identity_id", table_name="servednoticehistory"
    )
    op.drop_index(
        "ix_servednoticehistory_purpose_consent", table_name="servednoticehistory"
    )
    op.drop_index(
        "ix_servednoticehistory_purpose_legitimate_interests",
        table_name="servednoticehistory",
    )
    op.drop_index(
        "ix_servednoticehistory_special_feature", table_name="servednoticehistory"
    )
    op.drop_index(
        "ix_servednoticehistory_special_purpose", table_name="servednoticehistory"
    )
    op.drop_index(
        "ix_servednoticehistory_system_consent", table_name="servednoticehistory"
    )
    op.drop_index(
        "ix_servednoticehistory_system_legitimate_interests",
        table_name="servednoticehistory",
    )
    op.drop_index(
        "ix_servednoticehistory_vendor_consent", table_name="servednoticehistory"
    )
    op.drop_index(
        "ix_servednoticehistory_vendor_legitimate_interests",
        table_name="servednoticehistory",
    )
    op.create_index(
        op.f("ix_servednoticehistory_notice_key"),
        "servednoticehistory",
        ["notice_key"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_notice_mechanism"),
        "servednoticehistory",
        ["notice_mechanism"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_notice_name"),
        "servednoticehistory",
        ["notice_name"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_served_notice_history_id"),
        "servednoticehistory",
        ["served_notice_history_id"],
        unique=False,
    )
    op.drop_constraint(
        "servednoticehistory_provided_identity_id_fkey",
        "servednoticehistory",
        type_="foreignkey",
    )
    op.drop_constraint(
        "servednoticehistory_fides_user_device_provided_identity_id_fkey",
        "servednoticehistory",
        type_="foreignkey",
    )
    op.drop_column("servednoticehistory", "vendor_consent")
    op.drop_column("servednoticehistory", "purpose_consent")
    op.drop_column("servednoticehistory", "purpose_legitimate_interests")
    op.drop_column("servednoticehistory", "system_legitimate_interests")
    op.drop_column("servednoticehistory", "fides_user_device_provided_identity_id")
    op.drop_column("servednoticehistory", "special_feature")
    op.drop_column("servednoticehistory", "special_purpose")
    op.drop_column("servednoticehistory", "tcf_version")
    op.drop_column("servednoticehistory", "system_consent")
    op.drop_column("servednoticehistory", "vendor_legitimate_interests")
    op.drop_column("servednoticehistory", "feature")
    op.drop_column("servednoticehistory", "provided_identity_id")


def downgrade():
    op.alter_column(
        "privacypreferencehistory",
        "preference",
        type_=sa.Enum(ConsentMechanism),
        existing_type=sa.String(),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "provided_identity_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("feature", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "vendor_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("system_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("tcf_version", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("special_purpose", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("special_feature", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "fides_user_device_provided_identity_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "system_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "purpose_legitimate_interests",
            sa.INTEGER(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("purpose_consent", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("vendor_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_foreign_key(
        "servednoticehistory_fides_user_device_provided_identity_id_fkey",
        "servednoticehistory",
        "providedidentity",
        ["fides_user_device_provided_identity_id"],
        ["id"],
    )
    op.create_foreign_key(
        "servednoticehistory_provided_identity_id_fkey",
        "servednoticehistory",
        "providedidentity",
        ["provided_identity_id"],
        ["id"],
    )
    op.drop_index(
        op.f("ix_servednoticehistory_served_notice_history_id"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_notice_name"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_notice_mechanism"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_notice_key"), table_name="servednoticehistory"
    )
    op.create_index(
        "ix_servednoticehistory_vendor_legitimate_interests",
        "servednoticehistory",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_vendor_consent",
        "servednoticehistory",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_system_legitimate_interests",
        "servednoticehistory",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_system_consent",
        "servednoticehistory",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_special_purpose",
        "servednoticehistory",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_special_feature",
        "servednoticehistory",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_purpose_legitimate_interests",
        "servednoticehistory",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_purpose_consent",
        "servednoticehistory",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_provided_identity_id",
        "servednoticehistory",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_fides_user_device_provided_identity_id",
        "servednoticehistory",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_servednoticehistory_feature",
        "servednoticehistory",
        ["feature"],
        unique=False,
    )
    op.drop_column("servednoticehistory", "tcf_served")
    op.drop_column("servednoticehistory", "served_notice_history_id")
    op.drop_column("servednoticehistory", "notice_name")
    op.drop_column("servednoticehistory", "notice_mechanism")
    op.drop_column("servednoticehistory", "notice_key")
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "provided_identity_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("feature", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "vendor_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "relevant_systems",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("system_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("tcf_version", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("special_purpose", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("special_feature", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "fides_user_device_provided_identity_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "system_legitimate_interests",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "purpose_legitimate_interests",
            sa.INTEGER(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("purpose_consent", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("vendor_consent", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_foreign_key(
        "privacy_preference_served_notice_fk",
        "privacypreferencehistory",
        "servednoticehistory",
        ["served_notice_history_id"],
        ["id"],
    )
    op.create_foreign_key(
        "privacypreferencehistory_fides_user_device_provided_identi_fkey",
        "privacypreferencehistory",
        "providedidentity",
        ["fides_user_device_provided_identity_id"],
        ["id"],
    )
    op.create_foreign_key(
        "privacypreferencehistory_provided_identity_id_fkey",
        "privacypreferencehistory",
        "providedidentity",
        ["provided_identity_id"],
        ["id"],
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_notice_name"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_notice_mechanism"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_notice_key"),
        table_name="privacypreferencehistory",
    )
    op.create_index(
        "ix_privacypreferencehistory_vendor_legitimate_interests",
        "privacypreferencehistory",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_vendor_consent",
        "privacypreferencehistory",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_system_legitimate_interests",
        "privacypreferencehistory",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_system_consent",
        "privacypreferencehistory",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_special_purpose",
        "privacypreferencehistory",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_special_feature",
        "privacypreferencehistory",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_purpose_legitimate_interests",
        "privacypreferencehistory",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_purpose_consent",
        "privacypreferencehistory",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_provided_identity_id",
        "privacypreferencehistory",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_fides_user_device_provided__9961",
        "privacypreferencehistory",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        "ix_privacypreferencehistory_feature",
        "privacypreferencehistory",
        ["feature"],
        unique=False,
    )
    op.drop_column("privacypreferencehistory", "tcf_preferences")
    op.drop_column("privacypreferencehistory", "fides_string")
    op.drop_column("privacypreferencehistory", "notice_name")
    op.drop_column("privacypreferencehistory", "notice_mechanism")
    op.drop_column("privacypreferencehistory", "notice_key")
    op.add_column(
        "lastservednotice",
        sa.Column(
            "served_notice_history_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.create_index(
        "ix_lastservednotice_served_notice_history_id",
        "lastservednotice",
        ["served_notice_history_id"],
        unique=False,
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column(
            "privacy_preference_history_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.create_index(
        "ix_currentprivacypreference_privacy_preference_history_id",
        "currentprivacypreference",
        ["privacy_preference_history_id"],
        unique=False,
    )
    op.create_index("ix_ctl_systems_name", "ctl_systems", ["name"], unique=False)
    op.drop_index(
        op.f("ix_lastservednoticev2_updated_at"), table_name="lastservednoticev2"
    )
    op.drop_index(op.f("ix_lastservednoticev2_id"), table_name="lastservednoticev2")
    op.drop_index(
        op.f("ix_lastservednoticev2_hashed_phone_number"),
        table_name="lastservednoticev2",
    )
    op.drop_index(
        op.f("ix_lastservednoticev2_hashed_fides_user_device"),
        table_name="lastservednoticev2",
    )
    op.drop_index(
        op.f("ix_lastservednoticev2_hashed_email"), table_name="lastservednoticev2"
    )
    op.drop_index(
        op.f("ix_lastservednoticev2_created_at"), table_name="lastservednoticev2"
    )
    op.drop_table("lastservednoticev2")
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_updated_at"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_id"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_hashed_phone_number"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_hashed_fides_user_device"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_hashed_email"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_index(
        op.f("ix_currentprivacypreferencev2_created_at"),
        table_name="currentprivacypreferencev2",
    )
    op.drop_table("currentprivacypreferencev2")



"""
Creates staged resource ancestor link table.

Does not populate the table with data, that is handled in a separate data migration.

Also does not drop the StagedResource.child_diff_statuses column,
that is handled in a separate migration, _after_ the data migration
that populates the staged resource ancestor link table, to prevent
unintended data loss.

Revision ID: 5474a47c77da
Revises: 440a5b9a3493
Create Date: 2025-05-27 19:02:43.802783

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "5474a47c77da"
down_revision = "440a5b9a3493"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "stagedresourceancestor",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("ancestor_urn", sa.String(), nullable=False),
        sa.Column("descendant_urn", sa.String(), nullable=False),
        # primary key, foreign key and unique constraints are deferred
        # until after the data migration that populates the table,
        # to avoid slowing down the data migration
    )

    # defer creating `stagedresourceancestor` indexes until after the data migration
    # that populates the table, to avoid slowing down the data migration

    # add an index for StagedResource.diff_status to improve queries against the diff_status field
    op.create_index(
        op.f("ix_stagedresource_diff_status"),
        "stagedresource",
        ["diff_status"],
        unique=False,
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # drop the StagedResource.diff_status index created in the migration
    op.drop_index(op.f("ix_stagedresource_diff_status"), table_name="stagedresource")

    # drop the stagedresourceancestor table (and all its indexes and constraints)
    op.drop_table("stagedresourceancestor")
    # ### end Alembic commands ###



"""add notice and exp translations

Revision ID: a1e23b70f2b2
Revises: 26d5976531d6
Create Date: 2024-02-01 21:49:20.792733

"""

from datetime import datetime

import sqlalchemy as sa
from alembic import op
from loguru import logger
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "a1e23b70f2b2"
down_revision = "26d5976531d6"
branch_labels = None
depends_on = None


def upgrade():
    logger.info(
        f"Starting Consent Multitranslation Schema Migration (#1) {datetime.now()}"
    )

    op.create_table(
        "experienceconfigtemplate",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("regions", postgresql.ARRAY(sa.String()), nullable=True),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("component", sa.String(), nullable=False),
        sa.Column("privacy_notice_keys", postgresql.ARRAY(sa.String()), nullable=True),
        sa.Column(
            "translations",
            postgresql.ARRAY(postgresql.JSONB(astext_type=sa.Text())),
            nullable=True,
        ),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("dismissable", sa.Boolean(), server_default="t", nullable=False),
        sa.Column(
            "auto_detect_language", sa.Boolean(), server_default="t", nullable=False
        ),
        sa.Column(
            "allow_language_selection", sa.Boolean(), server_default="f", nullable=False
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_experienceconfigtemplate_id"),
        "experienceconfigtemplate",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_experienceconfigtemplate_component"),
        "experienceconfigtemplate",
        ["component"],
        unique=False,
    )
    op.create_table(
        "experiencenotices",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("id", sa.String(length=255), nullable=True),
        sa.Column("notice_id", sa.String(), nullable=False),
        sa.Column("experience_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["experience_config_id"], ["privacyexperienceconfig.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["notice_id"], ["privacynotice.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("notice_id", "experience_config_id"),
    )
    op.create_index(
        op.f("ix_experiencenotices_experience_config_id"),
        "experiencenotices",
        ["experience_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_experiencenotices_notice_id"),
        "experiencenotices",
        ["notice_id"],
        unique=False,
    )
    op.create_table(
        "experiencetranslation",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("language", sa.String(), nullable=False),
        sa.Column("accept_button_label", sa.String(), nullable=True),
        sa.Column("acknowledge_button_label", sa.String(), nullable=True),
        sa.Column("banner_description", sa.String(), nullable=True),
        sa.Column("banner_title", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("is_default", sa.Boolean(), nullable=False),
        sa.Column("privacy_policy_link_label", sa.String(), nullable=True),
        sa.Column("privacy_policy_url", sa.String(), nullable=True),
        sa.Column("privacy_preferences_link_label", sa.String(), nullable=True),
        sa.Column("reject_button_label", sa.String(), nullable=True),
        sa.Column("save_button_label", sa.String(), nullable=True),
        sa.Column("title", sa.String(), nullable=True),
        sa.Column("experience_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["experience_config_id"],
            ["privacyexperienceconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "language", "experience_config_id", name="experience_translation"
        ),
    )
    op.create_index(
        op.f("ix_experiencetranslation_id"),
        "experiencetranslation",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_experiencetranslation_experience_config_id"),
        "experiencetranslation",
        ["experience_config_id"],
        unique=False,
    )

    op.create_table(
        "noticetranslation",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("language", sa.String(), nullable=False),
        sa.Column("title", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("privacy_notice_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_notice_id"],
            ["privacynotice.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("language", "privacy_notice_id", name="notice_translation"),
    )
    op.create_index(
        op.f("ix_noticetranslation_id"), "noticetranslation", ["id"], unique=False
    )
    op.drop_constraint("region_component", "privacyexperience", type_="unique")
    op.drop_column("privacyexperience", "component")
    op.add_column(
        "privacyexperienceconfig", sa.Column("origin", sa.String(), nullable=True)
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("dismissable", sa.Boolean(), server_default="t", nullable=False),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "allow_language_selection", sa.Boolean(), server_default="f", nullable=False
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "auto_detect_language", sa.Boolean(), server_default="t", nullable=False
        ),
    )
    op.add_column(
        "privacyexperienceconfig", sa.Column("name", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_privacyexperienceconfig_disabled"),
        "privacyexperienceconfig",
        ["disabled"],
        unique=False,
    )
    op.create_foreign_key(
        "config_template_origin",
        "privacyexperienceconfig",
        "experienceconfigtemplate",
        ["origin"],
        ["id"],
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("language", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory", sa.Column("name", sa.String(), nullable=True)
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("origin", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("dismissable", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("auto_detect_language", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("allow_language_selection", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("translation_id", sa.String(), nullable=True),
    )
    op.create_index(
        op.f("ix_privacyexperienceconfighistory_translation_id"),
        "privacyexperienceconfighistory",
        ["translation_id"],
        unique=False,
    )
    op.alter_column(
        "privacyexperienceconfighistory",
        "experience_config_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    op.create_foreign_key(
        "confighistory_origin",
        "privacyexperienceconfighistory",
        "experienceconfigtemplate",
        ["origin"],
        ["id"],
    )
    op.create_foreign_key(
        "confighistory_translation_id",
        "privacyexperienceconfighistory",
        "experiencetranslation",
        ["translation_id"],
        ["id"],
        ondelete="SET NULL",
    )
    # drop FK constraint between PrivacyExperienceConfigHistory and PrivacyExperienceConfig,
    # to allow us to delete PrivacyExperienceConfigs without removing their associated history records
    op.drop_constraint(
        "privacyexperienceconfighistory_experience_config_id_fkey",
        "privacyexperienceconfighistory",
        type_="foreignkey",
    )
    op.alter_column(
        "privacynotice",
        "regions",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
    )
    op.add_column(
        "privacynoticehistory", sa.Column("language", sa.String(), nullable=True)
    )
    op.add_column(
        "privacynoticehistory",
        sa.Column(
            "title", sa.String(), nullable=True
        ),  # temporarily nullable for existing values
    )
    op.add_column(
        "privacynoticehistory", sa.Column("translation_id", sa.String(), nullable=True)
    )
    op.alter_column(
        "privacynoticehistory",
        "regions",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
    )
    op.alter_column(
        "privacynoticehistory",
        "displayed_in_privacy_center",
        existing_type=sa.BOOLEAN(),
        nullable=True,
    )
    op.alter_column(
        "privacynoticehistory",
        "displayed_in_overlay",
        existing_type=sa.BOOLEAN(),
        nullable=True,
    )
    op.alter_column(
        "privacynoticehistory",
        "displayed_in_api",
        existing_type=sa.BOOLEAN(),
        nullable=True,
    )
    op.alter_column(
        "privacynoticehistory",
        "privacy_notice_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    op.create_foreign_key(
        "noticehistory_translation_id",
        "privacynoticehistory",
        "noticetranslation",
        ["translation_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.create_index(
        op.f("ix_privacynoticehistory_translation_id"),
        "privacynoticehistory",
        ["translation_id"],
        unique=False,
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column(
            "translations",
            postgresql.ARRAY(postgresql.JSONB(astext_type=sa.Text())),
            nullable=True,
        ),
    )
    op.alter_column(
        "privacynoticetemplate",
        "regions",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
    )
    op.drop_index(
        "ix_privacypreferencehistory_privacy_experience_id",
        table_name="privacypreferencehistory",
    )
    op.drop_constraint(
        "privacypreferencehistory_privacy_experience_id_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("language", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_language"),
        "privacypreferencehistory",
        ["language"],
        unique=False,
    )
    op.add_column(
        "servednoticehistory", sa.Column("language", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_servednoticehistory_language"),
        "servednoticehistory",
        ["language"],
        unique=False,
    )
    op.drop_index(
        "ix_servednoticehistory_privacy_experience_id", table_name="servednoticehistory"
    )
    op.drop_constraint(
        "servednoticehistory_privacy_experience_id_fkey",
        "servednoticehistory",
        type_="foreignkey",
    )


def downgrade():
    op.create_foreign_key(
        "servednoticehistory_privacy_experience_id_fkey",
        "servednoticehistory",
        "privacyexperience",
        ["privacy_experience_id"],
        ["id"],
    )
    op.create_index(
        "ix_servednoticehistory_privacy_experience_id",
        "servednoticehistory",
        ["privacy_experience_id"],
        unique=False,
    )
    op.drop_index(
        op.f("ix_servednoticehistory_language"), table_name="servednoticehistory"
    )
    op.drop_column("servednoticehistory", "language")
    op.drop_index(
        op.f("ix_privacypreferencehistory_language"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfighistory_translation_id"),
        table_name="privacyexperienceconfighistory",
    )
    op.drop_column("privacypreferencehistory", "language")
    op.create_foreign_key(
        "privacypreferencehistory_privacy_experience_id_fkey",
        "privacypreferencehistory",
        "privacyexperience",
        ["privacy_experience_id"],
        ["id"],
    )
    op.create_index(
        "ix_privacypreferencehistory_privacy_experience_id",
        "privacypreferencehistory",
        ["privacy_experience_id"],
        unique=False,
    )
    op.alter_column(
        "privacynoticetemplate",
        "regions",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
    )
    op.drop_column("privacynoticetemplate", "translations")
    op.drop_constraint(
        "noticehistory_translation_id", "privacynoticehistory", type_="foreignkey"
    )
    op.drop_index(
        op.f("ix_privacynoticehistory_translation_id"),
        table_name="privacynoticehistory",
    )
    op.alter_column(
        "privacynoticehistory",
        "privacy_notice_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    op.alter_column(
        "privacynoticehistory",
        "regions",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
    )
    op.drop_column("privacynoticehistory", "translation_id")
    op.drop_column("privacynoticehistory", "title")
    op.drop_column("privacynoticehistory", "language")
    op.alter_column(
        "privacynotice",
        "regions",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
    )
    op.drop_constraint(
        "confighistory_translation_id",
        "privacyexperienceconfighistory",
        type_="foreignkey",
    )
    op.drop_constraint(
        "confighistory_origin", "privacyexperienceconfighistory", type_="foreignkey"
    )
    op.alter_column(
        "privacyexperienceconfighistory",
        "experience_config_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    op.drop_column("privacyexperienceconfighistory", "translation_id")
    op.drop_column("privacyexperienceconfighistory", "allow_language_selection")
    op.drop_column("privacyexperienceconfighistory", "auto_detect_language")
    op.drop_column("privacyexperienceconfighistory", "dismissable")
    op.drop_column("privacyexperienceconfighistory", "origin")
    op.drop_column("privacyexperienceconfighistory", "name")
    op.drop_column("privacyexperienceconfighistory", "language")

    op.drop_constraint(
        "config_template_origin", "privacyexperienceconfig", type_="foreignkey"
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfig_disabled"),
        table_name="privacyexperienceconfig",
    )
    op.drop_column("privacyexperienceconfig", "name")
    op.drop_column("privacyexperienceconfig", "allow_language_selection")
    op.drop_column("privacyexperienceconfig", "dismissable")
    op.drop_column("privacyexperienceconfig", "auto_detect_language")
    op.drop_column("privacyexperienceconfig", "origin")
    op.add_column(
        "privacyexperience",
        sa.Column("component", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_unique_constraint(
        "region_component", "privacyexperience", ["region", "component"]
    )
    op.drop_index(op.f("ix_noticetranslation_id"), table_name="noticetranslation")
    op.drop_table("noticetranslation")
    op.drop_index(
        op.f("ix_experiencetranslation_id"), table_name="experiencetranslation"
    )
    op.drop_index(
        op.f("ix_experiencetranslation_experience_config_id"),
        table_name="experiencetranslation",
    )
    op.drop_table("experiencetranslation")
    op.drop_index(
        op.f("ix_experiencenotices_notice_id"), table_name="experiencenotices"
    )
    op.drop_index(
        op.f("ix_experiencenotices_experience_config_id"),
        table_name="experiencenotices",
    )
    op.drop_table("experiencenotices")
    op.drop_index(
        op.f("ix_experienceconfigtemplate_id"), table_name="experienceconfigtemplate"
    )
    op.drop_table("experienceconfigtemplate")



from typing import Any, Dict, List, Optional, TypeVar

from boto3.dynamodb.types import TypeSerializer

from fides.api.graph.execution import ExecutionNode
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row

T = TypeVar("T")

DynamoDBStatement = Dict[str, Any]
"""A DynamoDB query is formed using the boto3 library. The required parameters are:
  * a table/collection name (string)
  * the key name to pass when accessing the table, along with type and value (dict)
  * optionally, the sort key or secondary index (i.e. timestamp)
  * optionally, the specified attributes can be provided. If None, all attributes
  returned for item.

    # TODO finish these docs

  We can either represent these items as a model and then handle each of the values
  accordingly in the connector or use this query config to return a dictionary that
  can be appropriately unpacked when executing using the client.

  The get_item query has been left out of the query_config for now.

  Add an example for put_item
  """


class DynamoDBQueryConfig(QueryConfig[DynamoDBStatement]):
    def __init__(
        self, node: ExecutionNode, attribute_definitions: List[Dict[str, Any]]
    ):
        super().__init__(node)
        self.attribute_definitions = attribute_definitions

    def generate_query(
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy],
    ) -> Optional[DynamoDBStatement]:
        """Generates a dictionary for the `query` method used for DynamoDB"""
        query_param = {}
        serializer = TypeSerializer()
        for attribute_definition in self.attribute_definitions:
            attribute_name = attribute_definition["AttributeName"]
            attribute_value = input_data[attribute_name][0]
            query_param["ExpressionAttributeValues"] = {
                ":value": serializer.serialize(attribute_value)
            }
            key_condition_expression: str = f"{attribute_name} = :value"
            query_param["KeyConditionExpression"] = key_condition_expression  # type: ignore
        return query_param

    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Optional[DynamoDBStatement]:
        """
        Generate a Dictionary that contains necessary items to
        run a PUT operation against DynamoDB
        """
        update_clauses = self.update_value_map(row, policy, request)

        if update_clauses:
            serializer = TypeSerializer()
            update_items = row
            for key, value in update_items.items():
                if key in update_clauses:
                    update_items[key] = serializer.serialize(update_clauses[key])
                else:
                    update_items[key] = serializer.serialize(value)
        else:
            update_items = None

        return update_items

    def query_to_str(self, t: T, input_data: Dict[str, List[Any]]) -> None:
        """Not used for this connector"""
        return None

    def dry_run_query(self) -> None:
        """Not used for this connector"""
        return None



from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import BaseModel, Field, ValidationError, model_validator
from sqlalchemy import Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, Index, Integer, String, insert, select, update
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base, FidesBase

if TYPE_CHECKING:
    from fides.api.models.privacy_experience import PrivacyExperienceConfig


class TCFRestrictionType(str, Enum):
    """Enum for TCF restriction types"""

    purpose_restriction = "purpose_restriction"
    require_consent = "require_consent"
    require_legitimate_interest = "require_legitimate_interest"


class TCFVendorRestriction(str, Enum):
    """Enum for TCF vendor restriction types"""

    restrict_all_vendors = "restrict_all_vendors"
    allow_specific_vendors = "allow_specific_vendors"
    restrict_specific_vendors = "restrict_specific_vendors"


# This is Pydantic model used for validation, not a database model!
class RangeEntry(BaseModel):
    """
    Pydantic model that represents a vendor range entry as per the TCF spec,
    used for Publisher Restrictions.
    A range entry must have a start_vendor_id and optionally an end_vendor_id.
    If end_vendor_id is present, it must be greater than start_vendor_id.
    """

    start_vendor_id: int = Field(description="The starting vendor ID in the range")
    end_vendor_id: Optional[int] = Field(
        default=None, description="The ending vendor ID in the range (inclusive)"
    )

    @model_validator(mode="after")
    def validate_vendor_range(self) -> "RangeEntry":
        """Validates that end_vendor_id is greater than start_vendor_id if present."""
        if self.end_vendor_id is not None and self.end_vendor_id < self.start_vendor_id:
            raise ValueError(
                "end_vendor_id must be greater than or equal to start_vendor_id"
            )
        return self

    @property
    def effective_end_vendor_id(self) -> int:
        """Get the effective end of the range."""
        return self.end_vendor_id or self.start_vendor_id

    def overlaps_with(self, other: "RangeEntry") -> bool:
        """
        Check if this range overlaps with another range.
        Two ranges overlap if the end of the range that starts first is greater than
        or equal to the start of the range that starts second.
        """
        # Sort ranges by start_vendor_id
        first = self if self.start_vendor_id <= other.start_vendor_id else other
        second = other if self.start_vendor_id <= other.start_vendor_id else self

        # If first range's end is >= second range's start, they overlap
        return first.effective_end_vendor_id >= second.start_vendor_id


class TCFConfiguration(Base):
    """
    Stores TCF Configuration settings.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "tcf_configuration"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    name = Column(String, nullable=False, unique=True)

    privacy_experience_configs = relationship(
        "PrivacyExperienceConfig",
        back_populates="tcf_configuration",
        lazy="selectin",
        viewonly=True,
    )


class TCFPublisherRestriction(Base):
    """
    Stores TCF Publisher Restrictions. TCF Publisher Restrictions belong to a TCF Configuration,
    and specify the restriction type and vendor restriction for a given purpose.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "tcf_publisher_restriction"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)

    tcf_configuration_id = Column(
        String(255),
        ForeignKey("tcf_configuration.id", ondelete="CASCADE"),
        nullable=False,
    )
    purpose_id = Column(Integer, nullable=False)
    restriction_type = Column(EnumColumn(TCFRestrictionType), nullable=False)
    vendor_restriction = Column(EnumColumn(TCFVendorRestriction), nullable=False)

    # range_entries represents a list of RangeEntry objects as per the TCF spec.
    # A RangeEntry object will have a start_vendor_id and an end_vendor_id (the
    # end vendor id is included in the range).
    # If the range represents a single vendor, then the end_vendor_id can be omitted.
    # TCF spec also includes an IsARange boolean, which we are omitting because it
    # can be inferred from the presence of the end_vendor_id.
    range_entries = Column(
        ARRAY(JSONB),
        nullable=False,
        server_default="{}",
        default=list,
    )

    __table_args__ = (
        Index(
            "ix_tcf_publisher_restriction_config_purpose",
            "tcf_configuration_id",
            "purpose_id",
        ),
    )

    def __init__(self, **kwargs: Any) -> None:
        validated_kwargs = self.validate_publisher_restriction_data(kwargs)
        super().__init__(**validated_kwargs)

    @staticmethod
    def validate_entires_for_vendor_restriction(
        entries: List[RangeEntry], vendor_restriction: TCFVendorRestriction
    ) -> None:
        """
        Validates that if vendor_restriction is restrict_all_vendors, then the entries list is empty.
        If vendor_restriction is not restrict_all_vendors, then there must be at least one entry.
        """
        if vendor_restriction == TCFVendorRestriction.restrict_all_vendors and entries:
            raise ValueError(
                "A restrict_all_vendors restriction cannot have any range entries"
            )

        if (
            vendor_restriction != TCFVendorRestriction.restrict_all_vendors
            and not entries
        ):
            raise ValueError(
                f"A {vendor_restriction} restriction must have at least one range entry"
            )

    @staticmethod
    def check_for_overlaps(entries: List[RangeEntry]) -> None:
        """
        Check for overlapping ranges in a list of RangeEntry objects,
        raises a ValueError if any pair of RangeEntry overlaps with each other.
        Sorts the ranges by start_vendor_id and compares adjacent ranges.
        """
        # Sort ranges by start_vendor_id
        sorted_entries = sorted(entries, key=lambda x: x.start_vendor_id)

        # Compare each range with the next one
        for i in range(len(sorted_entries) - 1):
            current = sorted_entries[i]
            next_entry = sorted_entries[i + 1]
            if current.overlaps_with(next_entry):
                raise ValueError(
                    f"Overlapping ranges found: {current.model_dump()} overlaps with {next_entry.model_dump()}"
                )

    @classmethod
    def validate_publisher_restriction_data(
        cls, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate the restriction data.
        """
        raw_range_entries = data.get("range_entries", [])

        try:
            # Validate each range entry using the Pydantic model
            validated_entries = [
                RangeEntry.model_validate(entry) for entry in raw_range_entries
            ]
        except ValidationError as e:
            raise ValueError(f"Invalid range entry: {str(e)}")

        # Validate the entries for the vendor restriction
        cls.validate_entires_for_vendor_restriction(
            validated_entries, data["vendor_restriction"]
        )
        # Check for overlapping ranges
        cls.check_for_overlaps(validated_entries)

        data["range_entries"] = [entry.model_dump() for entry in validated_entries]

        return data

    @classmethod
    async def check_for_restriction_conflicts(
        cls,
        *,
        async_db: AsyncSession,
        configuration_id: str,
        purpose_id: int,
        restriction_type: TCFRestrictionType,
        vendor_restriction: TCFVendorRestriction,
        range_entries: Optional[list[RangeEntry]] = None,
        restriction_id: Optional[str] = None,
    ) -> None:
        """
        Checks that the new restriction data does not conflict with any existing restrictions for the purpose.
        """
        # First, get all the restrictions for the purpose in the given configuration
        query = (
            select(cls)  # type: ignore[arg-type]
            .where(cls.tcf_configuration_id == configuration_id)
            .where(cls.purpose_id == purpose_id)
        )
        # If we're updating an existing restriction, exclude it from the list of
        # restrictions so it doesn't conflict with itself
        if restriction_id:
            query = query.where(cls.id != restriction_id)

        restrictions_result = await async_db.execute(query)
        relevant_restrictions = restrictions_result.scalars().all()

        if any(
            restriction.vendor_restriction == TCFVendorRestriction.restrict_all_vendors
            for restriction in relevant_restrictions
        ):
            raise ValueError(
                f"Invalid restriction for purpose {purpose_id}: a restrict_all_vendors restriction exists for this purpose."
            )

        # If we're creating a restrict_all_vendors restriction,
        # then there should be no other restrictions for this purpose
        if vendor_restriction == TCFVendorRestriction.restrict_all_vendors:
            if relevant_restrictions:
                raise ValueError(
                    f"Invalid restrict_all_vendors restriction for purpose {purpose_id}: other restrictions already exist for this purpose."
                )

            return None

        # If we already have a restriction for that restriction type,
        # then we raise an error
        if any(
            restriction.restriction_type == restriction_type
            for restriction in relevant_restrictions
        ):
            raise ValueError(
                f"Invalid {restriction_type} restriction for purpose {purpose_id}: a restriction of this type already exists for this purpose."
            )

        # We now need to check for vendor overlaps between all the restrictions.
        # To achieve this, we need to transform allowlist-style restrictions into
        # actual restriction ranges (rather than "allow" ranges).
        new_range_entries = (
            cls.transform_allowlist_restriction(range_entries or [])
            if vendor_restriction == TCFVendorRestriction.allow_specific_vendors
            else (range_entries or [])
        )

        existing_range_entries = []
        for restriction in relevant_restrictions:
            range_entries = [
                RangeEntry.model_validate(entry) for entry in restriction.range_entries
            ]
            transformed_range_entries = (
                cls.transform_allowlist_restriction(range_entries)
                if restriction.vendor_restriction
                == TCFVendorRestriction.allow_specific_vendors
                else range_entries
            )
            existing_range_entries.extend(transformed_range_entries)

        all_entries = [*existing_range_entries, *new_range_entries]
        cls.check_for_overlaps(all_entries)

    @classmethod
    def transform_allowlist_restriction(
        cls, range_entries: list[RangeEntry]
    ) -> list[RangeEntry]:
        """
        Transform allowlist-style restrictions into restriction ranges.
        E.g if you have an "allow specific vendors" restriction with the range_entries
        [
            {start_vendor_id: 5, end_vendor_id: 10},
            {start_vendor_id: 25, end_vendor_id: 40},
            {start_vendor_id: 123 },
            {start_vendor_id: 345, end_vendor_id: 380},
        ],
        the transformed restriction ranges would be:
        [
            {start_vendor_id: 1, end_vendor_id: 4},
            {start_vendor_id: 11, end_vendor_id: 24},
            {start_vendor_id: 41, end_vendor_id: 122},
            {start_vendor_id: 124, end_vendor_id: MAX_GVL_ID},
        ]
        """
        MAX_GVL_ID = 9999  # TODO: get this from the TCF spec

        # First, we need to sort the range_entries by start_vendor_id
        sorted_range_entries = sorted(range_entries, key=lambda x: x.start_vendor_id)

        # Now, we need to transform the allowlist-style restrictions into
        # actual restriction ranges.
        transformed_range_entries: list[RangeEntry] = []

        total_entries = len(sorted_range_entries)

        # This shouldn't happen, but just in case
        if total_entries == 0:
            raise ValueError(
                "No range entries found for allow_specific_vendors restriction"
            )

        # If the first range entry starts at a number greater than 1,
        # we need to add a transformed range entry from 1 up to the entry's start
        if sorted_range_entries[0].start_vendor_id > 1:
            transformed_range_entries.append(
                RangeEntry(
                    start_vendor_id=1,
                    end_vendor_id=sorted_range_entries[0].start_vendor_id - 1,
                )
            )

        # Iterate through the sorted range_entries and transform them into restriction ranges
        for idx, range_entry in enumerate(sorted_range_entries):
            # For all but the last range entry, we add an entry that corresponds to the numbers
            # between the end of the current range entry and the start of the next range entry
            if idx < total_entries - 1:
                # Only create a range entry if there's actually a gap between the ranges
                next_start = sorted_range_entries[idx + 1].start_vendor_id
                current_end = range_entry.effective_end_vendor_id
                if next_start > current_end + 1:
                    transformed_range_entries.append(
                        RangeEntry(
                            start_vendor_id=current_end + 1,
                            end_vendor_id=next_start - 1,
                        )
                    )

        # If the last range entry ends at a number less than MAX_GVL_ID,
        # we need to add a transformed range entry from the last entry's end to MAX_GVL_ID
        if sorted_range_entries[-1].effective_end_vendor_id < MAX_GVL_ID:
            transformed_range_entries.append(
                RangeEntry(
                    start_vendor_id=sorted_range_entries[-1].effective_end_vendor_id
                    + 1,
                    end_vendor_id=MAX_GVL_ID,
                )
            )

        # Return the transformed restriction entries
        return transformed_range_entries

    @classmethod
    def create(
        cls,
        db: Session,
        *,
        data: Dict[str, Any],
        check_name: bool = True,
    ) -> "TCFPublisherRestriction":
        raise NotImplementedError("Use create_async instead")

    @classmethod
    async def create_async(
        cls,
        async_db: AsyncSession,
        *,
        data: Dict[str, Any],
    ) -> "TCFPublisherRestriction":
        """
        Create a new TCFPublisherRestriction with validated range_entries.
        """

        data = cls.validate_publisher_restriction_data(data)

        await cls.check_for_restriction_conflicts(
            async_db=async_db,
            configuration_id=data["tcf_configuration_id"],
            purpose_id=data["purpose_id"],
            restriction_type=TCFRestrictionType(data["restriction_type"]),
            vendor_restriction=TCFVendorRestriction(data["vendor_restriction"]),
            range_entries=[
                RangeEntry.model_validate(entry) for entry in data["range_entries"]
            ],
        )

        values = {
            "tcf_configuration_id": data["tcf_configuration_id"],
            "purpose_id": data["purpose_id"],
            "restriction_type": data["restriction_type"],
            "vendor_restriction": data["vendor_restriction"],
            "range_entries": data["range_entries"],
        }

        # Insert the new restriction
        insert_stmt = insert(cls).values(values)  # type: ignore[arg-type]
        result = await async_db.execute(insert_stmt)
        record_id = result.inserted_primary_key.id

        created_record = await async_db.execute(select(cls).where(cls.id == record_id))  # type: ignore[arg-type]
        return created_record.scalars().first()

    async def update_async(
        self, async_db: AsyncSession, data: Dict[str, Any]
    ) -> "TCFPublisherRestriction":
        """
        Update a TCFPublisherRestriction with the data.
        Validates the data and checks for vendor overlaps.
        """
        # Create a new dict merging the existing data and the updated data
        updated_data = {
            "id": self.id,
            "tcf_configuration_id": self.tcf_configuration_id,
            "purpose_id": self.purpose_id,
            "restriction_type": self.restriction_type,
            "vendor_restriction": self.vendor_restriction,
            "range_entries": self.range_entries,
            **data,
        }

        # First validate the data on its own
        data = self.validate_publisher_restriction_data(updated_data)

        # Then check for conflicts
        await self.check_for_restriction_conflicts(
            async_db=async_db,
            configuration_id=self.tcf_configuration_id,
            purpose_id=self.purpose_id,
            restriction_type=TCFRestrictionType(data["restriction_type"]),
            vendor_restriction=TCFVendorRestriction(data["vendor_restriction"]),
            range_entries=[
                RangeEntry.model_validate(entry) for entry in data["range_entries"]
            ],
            restriction_id=self.id,
        )

        # Remove the id from the updated
        updated_data.pop("id")

        # Finally, make the update
        update_query = (
            update(TCFPublisherRestriction)  # type: ignore[arg-type]
            .where(TCFPublisherRestriction.id == self.id)
            .values(**updated_data)
        )
        await async_db.execute(update_query)
        await async_db.commit()
        await async_db.refresh(self)

        return self



from typing import Any, Dict

from sqlalchemy import Boolean, CheckConstraint, Column, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base, FidesBase

default_privacy_center_config = {
    "actions": [
        {
            "policy_key": "default_access_policy",
            "title": "Access your data",
            "identity_inputs": {"email": "required"},
        },
        {
            "policy_key": "default_erasure_policy",
            "title": "Erase your data",
            "identity_inputs": {"email": "required"},
        },
    ],
}


class PrivacyCenterConfig(Base):
    """
    A single-row table used to store the subset of the Privacy Center's config.json that is supported by the Admin UI.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "plus_privacy_center_config"

    config = Column(
        MutableDict.as_mutable(JSONB),
        nullable=False,
    )
    single_row = Column(
        Boolean,
        default=True,
        nullable=False,
        unique=True,
    )  # used to constrain table to one row

    CheckConstraint("single_row", name="plus_privacy_center_config_single_row_check")
    UniqueConstraint("single_row", name="plus_privacy_center_config_single_row_unique")

    @classmethod
    def create_or_update(  # type: ignore[override]
        cls,
        db: Session,
        *,
        data: Dict[str, Any],
    ) -> FidesBase:
        """
        Creates a new config record if one doesn't exist, or updates the existing record.

        Here we effectively prevent more than a single record in the table.
        """
        existing_record = db.query(cls).first()
        if existing_record:
            updated_record = existing_record.update(
                db=db,
                data=data,
            )  # type: ignore[arg-type]
            return updated_record

        return cls.create(db=db, data=data)



"""add connection config description

Revision ID: b3b68c87c4a0
Revises: c3472d75c80e
Create Date: 2022-06-13 17:24:56.889227

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b3b68c87c4a0"
down_revision = "c3472d75c80e"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "connectionconfig", sa.Column("description", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_connectionconfig_description"),
        "connectionconfig",
        ["description"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_connectionconfig_description"), table_name="connectionconfig"
    )
    op.drop_column("connectionconfig", "description")



"""add password login enabled and totp secret to fidesuser

Revision ID: 99c603c1b8f9
Revises: 6e565c16dae1
Create Date: 2025-04-02 01:55:57.890545

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "99c603c1b8f9"
down_revision = "6e565c16dae1"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "fidesuser",
        sa.Column(
            "password_login_enabled",
            sa.Boolean(),
            nullable=True,
        ),
    )
    op.add_column(
        "fidesuser",
        sa.Column(
            "totp_secret",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.alter_column("fidesuser", "hashed_password", nullable=True)
    op.alter_column("fidesuser", "salt", nullable=True)


def downgrade():
    op.alter_column("fidesuser", "hashed_password", nullable=False)
    op.alter_column("fidesuser", "salt", nullable=False)
    op.drop_column("fidesuser", "totp_secret")
    op.drop_column("fidesuser", "password_login_enabled")



from abc import abstractmethod
from typing import Any, Optional, Set, Tuple

from sqlalchemy import Boolean, Column, Index
from sqlalchemy.ext.declarative import declared_attr

from fides.api.migrations.hash_migration_tracker import HashMigrationTracker
from fides.api.schemas.redis_cache import MultiValue


class HashMigrationMixin:
    """
    A mixin that includes common logic for models that have hashed fields that need be be migrated from using bcrypt to SHA-256.
    """

    # set existing rows to false but set to true on create since we'll
    # be using a new hashing function moving forward
    is_hash_migrated = Column(Boolean, nullable=False, server_default="f", default=True)

    @declared_attr
    def __table_args__(cls: Any) -> Tuple:
        """
        Creates a dynamically named partial index on the table to track unmigrated rows.
        The size of the index will shrink as more rows are migrated.
        """
        return (
            Index(
                f"idx_{cls.__tablename__}_unmigrated",  # type: ignore[attr-defined]
                "is_hash_migrated",
                postgresql_where=cls.is_hash_migrated.is_(False),  # type: ignore[attr-defined]
            ),
        )

    @classmethod
    @abstractmethod
    def bcrypt_hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        """Hash value using bcrypt."""

    @classmethod
    @abstractmethod
    def hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        """Hash value using SHA-256."""

    @classmethod
    def hash_value_for_search(
        cls,
        value: MultiValue,
    ) -> Set[str]:
        """
        Returns the SHA-256 hash and the bcrypt hash if the hash(es) on this table haven't been migrated yet.
        Once the migration is complete for this model, the function will only return the faster,
        SHA-256 and avoid the computationally expensive bcrypt hash.
        """

        hashed_values = set()
        if hashed_value := cls.hash_value(value):
            hashed_values.add(hashed_value)
        if not HashMigrationTracker.is_migrated(cls.__name__):
            if hashed_value := cls.bcrypt_hash_value(value):
                hashed_values.add(hashed_value)
        return hashed_values

    @abstractmethod
    def migrate_hashed_fields(self) -> None:
        """
        Abstract method to migrate a single instance of a model with hashed fields.
        This method should be implemented by each model class that uses HashMigrationMixin.

        The implementation should:
        1. Read the encrypted fields
        2. Re-hash them with SHA-256
        3. Assign the new hash values to the hashed fields
        4. Mark the instance as migrated
        """



from typing import Any, Dict, List, Optional, Union, cast

import pydash
from fideslang.models import MaskingStrategies
from loguru import logger
from sqlalchemy import MetaData, Table, or_, text
from sqlalchemy.engine import Engine
from sqlalchemy.sql import Delete, Update
from sqlalchemy.sql.elements import ColumnElement, TextClause
from sqlalchemy_bigquery import BigQueryDialect

from fides.api.graph.config import Field, FieldPath
from fides.api.graph.execution import ExecutionNode
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.namespace_meta.bigquery_namespace_meta import (
    BigQueryNamespaceMeta,
)
from fides.api.schemas.partitioning import TIME_BASED_REQUIRED_KEYS, combine_partitions
from fides.api.schemas.partitioning.time_based_partitioning import TimeBasedPartitioning
from fides.api.service.connectors.query_configs.query_config import (
    QueryStringWithoutTuplesOverrideQueryConfig,
)
from fides.api.util.collection_util import (
    Row,
    filter_nonempty_values,
    flatten_dict,
    merge_dicts,
    replace_none_arrays,
    unflatten_dict,
)


class BigQueryQueryConfig(QueryStringWithoutTuplesOverrideQueryConfig):
    """
    Generates SQL valid for BigQuery
    """

    namespace_meta_schema = BigQueryNamespaceMeta

    @property
    def partitioning(
        self,
    ) -> Optional[Union[List[TimeBasedPartitioning], Dict[str, Any]]]:
        # Overridden from base implementation to allow for _only_ BQ partitioning, for now
        return self.node.collection.partitioning

    def get_partition_clauses(
        self,
    ) -> List[str]:
        """
        Build a list of SQL `WHERE` clause strings for the collection's partitioning
        configuration.

        Supported modes:
        1. `Legacy static` - A list of `where_clauses`.
        2. `Single time-based` - A single time-based partitioning spec
        3. `Multiple time-based` - A list of time-based partitioning specs, with different intervals

        Any other combination (e.g. mixing modes or missing keys) raises a
        `ValueError` so that mis-configurations surface early.
        """

        partition_spec: Optional[Union[List[TimeBasedPartitioning], Dict[str, Any]]] = (
            self.partitioning
        )
        if not partition_spec:
            logger.warning(
                f"No partitioning specification found for node '{self.node.address}', skipping partition clauses"
            )
            return []

        # Legacy mode using `where_clauses`
        if isinstance(partition_spec, dict) and "where_clauses" in partition_spec:
            if any(k in partition_spec for k in TIME_BASED_REQUIRED_KEYS):
                # Mixed mode not allowed
                raise ValueError(
                    "Partitioning spec cannot define both `where_clauses` and time-based partitioning."
                )
            where = partition_spec.get("where_clauses") or []
            if not where:
                raise ValueError("`where_clauses` must be a non-empty list.")
            return where

        # Combine and de-duplicate boundary rows
        combined_expressions = combine_partitions(partition_spec)  # type: ignore

        dialect = BigQueryDialect()
        where_clauses = [
            str(expr.compile(dialect=dialect, compile_kwargs={"literal_binds": True}))
            for expr in combined_expressions
        ]

        return where_clauses

    def _generate_table_name(self) -> str:
        """
        Prepends the dataset ID and project ID to the base table name
        if the BigQuery namespace meta is provided.
        """

        table_name = self.node.collection.name
        if self.namespace_meta:
            bigquery_namespace_meta = cast(BigQueryNamespaceMeta, self.namespace_meta)
            table_name = f"{bigquery_namespace_meta.dataset_id}.{table_name}"
            if project_id := bigquery_namespace_meta.project_id:
                table_name = f"{project_id}.{table_name}"
        return table_name

    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """
        Returns a query string with backtick formatting for tables that have the same names as
        BigQuery reserved words.
        """
        return f'SELECT {field_list} FROM `{self._generate_table_name()}` WHERE ({" OR ".join(clauses)})'

    def generate_masking_stmt(
        self,
        node: ExecutionNode,
        row: Row,
        policy: Policy,
        request: PrivacyRequest,
        client: Engine,
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> Union[List[Update], List[Delete]]:
        """
        Generate a masking statement for BigQuery.

        If a masking override is present, it will take precedence over the policy masking strategy.
        """

        masking_override = node.collection.masking_strategy_override
        if masking_override and masking_override.strategy == MaskingStrategies.DELETE:
            logger.info(
                f"Masking override detected for collection {node.address.value}: {masking_override.strategy.value}"
            )
            return self.generate_delete(client, input_data or {})
        return self.generate_update(row, policy, request, client)

    def generate_update(
        self, row: Row, policy: Policy, request: PrivacyRequest, client: Engine
    ) -> List[Update]:
        """
        Using TextClause to insert 'None' values into BigQuery throws an exception, so we use update clause instead.
        Returns a List of SQLAlchemy Update object. Does not actually execute the update object.

        A List of multiple Update objects are returned for partitioned tables; for a non-partitioned table,
        a single Update object is returned in a List for consistent typing.

        This implementation handles nested fields by grouping them as JSON objects rather than
        individual field updates.

        See the README.md in this directory for a detailed example of how nested data is handled.
        """

        # 1. Take update_value_map as-is (already flattened)
        update_value_map: Dict[str, Any] = self.update_value_map(row, policy, request)

        # 2. Flatten the row
        flattened_row = flatten_dict(row)

        # 3. Merge flattened_row with update_value_map (update_value_map takes precedence)
        merged_dict = merge_dicts(flattened_row, update_value_map)

        # 4. Unflatten the merged dictionary
        nested_result = unflatten_dict(merged_dict)

        # 5. Replace any arrays containing only None values with empty arrays
        nested_result = replace_none_arrays(nested_result)

        # 6. Only keep top-level keys that are in the update_value_map
        top_level_keys = {key.split(".")[0] for key in update_value_map}

        # Filter the nested result to only include those top-level keys
        final_update_map = {
            k: v for k, v in nested_result.items() if k in top_level_keys
        }

        # Use existing non-empty reference fields mechanism for WHERE clause
        non_empty_reference_field_keys: Dict[str, Field] = filter_nonempty_values(
            {
                fpath.string_path: fld.cast(pydash.get(row, fpath.string_path))
                for fpath, fld in self.reference_field_paths.items()
                if pydash.get(row, fpath.string_path) is not None
            }
        )

        valid = len(non_empty_reference_field_keys) > 0 and final_update_map
        if not valid:
            logger.warning(
                "There is not enough data to generate a valid update statement for {}",
                self.node.address,
            )
            return []

        table = Table(self._generate_table_name(), MetaData(bind=client), autoload=True)
        where_clauses: List[ColumnElement] = [
            table.c[k] == v for k, v in non_empty_reference_field_keys.items()
        ]

        # Create update values using Column objects as keys to handle column names with spaces
        update_values = {}
        for column_name, value in final_update_map.items():
            # Use bracket notation to access columns with spaces in their names
            column = table.c[column_name]
            update_values[column] = value

        if self.partitioning:
            partition_clauses = self.get_partition_clauses()
            partitioned_queries = []
            logger.info(
                f"Generating {len(partition_clauses)} partition queries for node '{self.node.address}' in DSR execution"
            )
            for partition_clause in partition_clauses:
                partitioned_queries.append(
                    table.update()
                    .where(*(where_clauses + [text(partition_clause)]))
                    .values(update_values)
                )

            return partitioned_queries

        return [table.update().where(*where_clauses).values(update_values)]

    def generate_delete(
        self,
        client: Engine,
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> List[Delete]:
        """
        Returns a List of SQLAlchemy DELETE statements for BigQuery. Does not actually execute the delete statement.

        Used when a collection-level masking override is present and the masking strategy is DELETE.

        A List of multiple DELETE statements are returned for partitioned tables; for a non-partitioned table,
        a single DELETE statement is returned in a List for consistent typing.
        """

        if not input_data:
            logger.warning(
                "No input data provided for node '{}', skipping DELETE statement generation",
                self.node.address,
            )
            return []

        filtered_data = self.node.typed_filtered_values(input_data)

        if not filtered_data:
            logger.warning(
                "There is not enough data to generate a valid DELETE statement for {}",
                self.node.address,
            )
            return []

        table = Table(self._generate_table_name(), MetaData(bind=client), autoload=True)

        # Build individual reference clauses
        where_clauses: List[ColumnElement] = []
        for column_name, values in filtered_data.items():
            if len(values) == 1:
                where_clauses.append(table.c[column_name] == values[0])
            else:
                where_clauses.append(table.c[column_name].in_(values))

        # Combine reference clauses with OR instead of AND
        combined_reference_clause = or_(*where_clauses)

        if self.partitioning:
            partition_clauses = self.get_partition_clauses()
            partitioned_queries = []
            logger.info(
                f"Generating {len(partition_clauses)} partition queries for node '{self.node.address}' in DSR execution"
            )

            for partition_clause in partition_clauses:
                partitioned_queries.append(
                    table.delete()
                    .where(combined_reference_clause)
                    .where(text(partition_clause))
                )

            return partitioned_queries

        return [table.delete().where(combined_reference_clause)]

    def uses_delete_masking_strategy(self) -> bool:
        """Check if this collection uses DELETE masking strategy.

        Returns True if masking override is present and strategy is DELETE.
        """
        masking_override = self.node.collection.masking_strategy_override
        return (
            masking_override is not None
            and masking_override.strategy == MaskingStrategies.DELETE
        )

    def format_fields_for_query(
        self,
        field_paths: List[FieldPath],
    ) -> List[str]:
        """
        Returns field paths in a format they can be added into SQL queries.
        Only returns non-nested fields (fields with exactly one level).
        """

        formatted_fields = []
        for field_path in field_paths:
            if len(field_path.levels) == 1:
                formatted_fields.append(field_path.levels[0])
        return formatted_fields

    def format_clause_for_query(
        self, string_path: str, operator: str, operand: str
    ) -> str:
        """
        Returns clauses with proper BigQuery backtick escaping for column names.
        Handles column names with spaces and nested fields (dot-separated) by escaping each part individually.
        """
        # For nested fields (containing dots), escape each part individually
        if "." in string_path:
            parts = string_path.split(".")
            escaped_field = ".".join(f"`{part}`" for part in parts)
        else:
            # For simple fields, wrap the entire name in backticks
            escaped_field = f"`{string_path}`"

        if operator == "IN":
            return f"{escaped_field} IN ({operand})"
        return f"{escaped_field} {operator} :{operand}"

    def generate_raw_query_without_tuples(
        self, field_list: List[str], filters: Dict[str, List[Any]]
    ) -> Optional[TextClause]:
        """
        Allows executing a somewhat raw query where the field_list and filters do not depend
        on the Node or Graph structure.

        This is an override of the base class method that supports nested fields for BigQuery.

        Examples with field names containing dots and spaces, notice these are replaced with underscores in the parameter bindings:

        1. Single value filter:
           field_list = ["id", "name", "email"]
           filters = {"user.id": [123]}

           Generates: SELECT id, name, email FROM `project_id.dataset_id.table_name` WHERE (`user`.`id` = :user_id)
           With parameter binding: user_id = 123

        2. Field with spaces:
           field_list = ["id", "custom id", "email"]
           filters = {"custom id": ["abc123"]}

           Generates: SELECT id, `custom id`, email FROM `project_id.dataset_id.table_name` WHERE (`custom id` = :custom_id)
           With parameter binding: custom_id = "abc123"

        3. Multiple value filter with nested field:
           field_list = ["id", "name", "email"]
           filters = {"contact_info.primary_email": ["active", "pending"]}

           Generates: SELECT id, name, email FROM `project_id.dataset_id.table_name` WHERE (`contact_info`.`primary_email` IN (:contact_info_primary_email_in_stmt_generated_0, :contact_info_primary_email_in_stmt_generated_1))
           With parameter bindings: contact_info_primary_email_in_stmt_generated_0 = "active", contact_info_primary_email_in_stmt_generated_1 = "pending"
        """
        clauses = []
        query_data = {}
        for field_name, field_value in filters.items():
            # Replace dots and spaces with underscores in field names for parameter binding
            # SQLAlchemy parameter names cannot contain spaces or special characters
            field_binding_name = field_name.replace(".", "_").replace(" ", "_")
            data = set(field_value)
            if len(data) == 1:
                clauses.append(
                    self.format_clause_for_query(field_name, "=", field_binding_name)
                )
                query_data[field_binding_name] = data.pop()
            elif len(data) > 1:
                data_vals = list(data)
                query_data_keys: List[str] = []
                for val in data_vals:
                    # appending "_in_stmt_generated_" (can be any arbitrary str) so that this name has lower chance of conflicting with pre-existing column in table
                    query_data_name = (
                        field_binding_name
                        + "_in_stmt_generated_"
                        + str(data_vals.index(val))
                    )
                    query_data[query_data_name] = val
                    query_data_keys.append(self.format_query_data_name(query_data_name))
                operand = ", ".join(query_data_keys)
                clauses.append(self.format_clause_for_query(field_name, "IN", operand))

        if len(clauses) > 0:
            formatted_fields = ", ".join([f"`{field}`" for field in field_list])
            query_str = self.get_formatted_query_string(formatted_fields, clauses)
            return text(query_str).params(query_data)

        return None



"""system dictionary data migration

Revision ID: fd52d5f08c17
Revises: 39397820a0d5
Create Date: 2023-08-04 14:14:32.421414

"""

import json
from typing import Any, Dict, List, Optional, Tuple

from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import TextClause

from fides.config import CONFIG

revision = "fd52d5f08c17"
down_revision = "39397820a0d5"
branch_labels = None
depends_on = None

JOINT_CONTROLLER_INFO = "joint_controller_info"
DOES_INTERNATIONAL_TRANSFERS = "does_international_transfers"
RESPONSIBILITY = "responsibility"
REQUIRES_DATA_PROTECTION_ASSESSMENTS = "requires_data_protection_assessments"
DPA_LOCATION = "dpa_location"
DPA_PROGRESS = "dpa_progress"

LEGAL_BASIS_FOR_PROCESSING = "legal_basis_for_processing"
IMPACT_ASSESSMENT_LOCATION = "impact_assessment_location"
PROCESSES_SPECIAL_CATEGORY_DATA = "processes_special_category_data"
SPECIAL_CATEGORY_LEGAL_BASIS = "special_category_legal_basis"
DATA_SHARED_WITH_THIRD_PARTIES = "data_shared_with_third_parties"
THIRD_PARTIES = "third_parties"
RETENTION_PERIOD = "retention_period"


def _system_transform_joint_controller_info(
    joint_controller_decrypted: Optional[Dict],
    datasets_joint_controller_decrypted: List[Optional[Dict]],
) -> Optional[str]:
    """Migrating systems.joint_controller and datasets.joint_controller encrypted json string fields to
    System.joint_controller_info which is just a string.

    Prioritizing system joint_controller info if it exists, or taking the first non-null entry from datasets
    otherwise.
    """
    new_joint_controller_info: Optional[str] = None
    # Prioritize system data over dataset data
    joint_controller_data = (
        joint_controller_decrypted
        or next(  # Multiple datasets can be related to a system, so we use the first non null one if applicable
            (
                item
                for item in datasets_joint_controller_decrypted or []
                if item is not None
            ),
            None,
        )
    )
    if joint_controller_data and isinstance(joint_controller_data, dict):
        separated_fields = [
            joint_controller_data.get("name"),
            joint_controller_data.get("address"),
            joint_controller_data.get("email"),
            joint_controller_data.get("phone"),
        ]

        # If data looks like {"name": "N/A", "address": "N/A"...}, return None instead of an empty string
        new_joint_controller_info = (
            "; ".join([elem for elem in separated_fields if elem and elem != "N/A"])
            or None
        )

    return new_joint_controller_info


def _system_calculate_does_international_transfers(
    system_third_country_transfers: Optional[List[str]],
    dataset_third_country_transfers: Optional[List[str]],
) -> bool:
    """Migrating System.third_country_transfers and Dataset.third_country_transfers to System.does_international_transfers
    If any countries are listed, regardless of which country it is, System.does_international_transfers will be set to true.
    """
    return any([system_third_country_transfers, dataset_third_country_transfers])


def _system_transform_data_responsibility_title_type(
    dept_str: Optional[str],
) -> List[str]:
    """Migrating System.data_responsibility_title (str) to System.responsibility(list).  There can now be potentially many
    items listed here"""
    return [dept_str] if dept_str else []


def _flatten_data_protection_impact_assessment(
    data_protection_impact_assessment: Optional[Dict],
) -> Tuple[bool, Optional[str], Optional[str]]:
    """Flattening data from System.data_protection_impact_assessment into three separate fields:
    System.requires_data_protection_assessments, System.dpa_location, and System.dpa_progress
    """
    if not data_protection_impact_assessment or not isinstance(
        data_protection_impact_assessment, dict
    ):
        return False, None, None

    requires_data_protection_assessments = data_protection_impact_assessment.get(
        "is_required", False
    )
    dpa_progress = data_protection_impact_assessment.get("progress")
    dpa_location = data_protection_impact_assessment.get("link")
    return requires_data_protection_assessments, dpa_progress, dpa_location


def system_dictionary_additive_migration(bind: Connection):
    """Copy and/or transform existing data in Datasets, PrivacyDeclarations and the System itself to new
    Systems fields to support dictionary work"""
    get_systems_query: TextClause = text(
        """
        SELECT 
            ctl_systems.id,
            pgp_sym_decrypt(ctl_systems.joint_controller::bytea, :encryption_key) AS joint_controller_decrypted,
            ctl_systems.third_country_transfers,
            ctl_systems.data_responsibility_title,
            ctl_systems.data_protection_impact_assessment,
            array_accum(ctl_datasets.third_country_transfers) as dataset_third_country_transfers,
            array_agg(pgp_sym_decrypt(ctl_datasets.joint_controller::bytea, :encryption_key)) AS datasets_joint_controller_decrypted
        FROM ctl_systems
        LEFT JOIN privacydeclaration ON ctl_systems.id = privacydeclaration.system_id 
        LEFT JOIN ctl_datasets ON ctl_datasets.fides_key = ANY(privacydeclaration.dataset_references) 
        GROUP BY ctl_systems.id
        """
    )
    for row in bind.execute(
        get_systems_query, {"encryption_key": CONFIG.user.encryption_key}
    ):
        (
            requires_data_protection_assessments,
            dpa_progress,
            dpa_location,
        ) = _flatten_data_protection_impact_assessment(
            row["data_protection_impact_assessment"]
        )

        system_joint_controller_data: Optional[Dict] = json.loads(
            row["joint_controller_decrypted"] or "null"
        )
        # Potentially multiple datasets on systems that have joint controller info
        datasets_joint_controller_data: List[Optional[Dict]] = [
            json.loads(item)
            for item in row["datasets_joint_controller_decrypted"] or "null"
            if item
        ]

        system_data = {
            JOINT_CONTROLLER_INFO: _system_transform_joint_controller_info(
                system_joint_controller_data, datasets_joint_controller_data
            ),  # System.joint_controller_info populated from flattened data from either System.joint_controller
            # or Datasets.joint_controller. New field is not as rigid
            DOES_INTERNATIONAL_TRANSFERS: _system_calculate_does_international_transfers(
                row["third_country_transfers"], row["dataset_third_country_transfers"]
            ),  # System.does_international_transfers is populated based on whether System.third_country_transfers
            # or Dataset.third_country_transfers exist
            RESPONSIBILITY: _system_transform_data_responsibility_title_type(
                row[
                    "data_responsibility_title"
                ]  # From System.data_responsibility_title
            ),
            REQUIRES_DATA_PROTECTION_ASSESSMENTS: requires_data_protection_assessments,  # From System.data_protection_impact_assessment
            DPA_LOCATION: dpa_location,  # From System.data_protection_impact_assessment
            DPA_PROGRESS: dpa_progress,  # From System.data_protection_impact_assessment
        }

        update_system_query: TextClause = text(
            """
                UPDATE ctl_systems
                SET 
                    joint_controller_info = :joint_controller_info,
                    does_international_transfers = :does_international_transfers,
                    responsibility = :responsibility,
                    requires_data_protection_assessments = :requires_data_protection_assessments,
                    dpa_location = :dpa_location,
                    dpa_progress = :dpa_progress
                WHERE id = :id;
            """
        )

        system_data: Dict[str, Any] = system_data
        bind.execute(
            update_system_query,
            {"id": row["id"], **system_data},
        )


# Casing is changing
legal_basis_for_processing_map: Dict = {
    "Consent": "Consent",
    "Contract": "Contract",
    "Legal Obligation": "Legal obligations",
    "Legal obligations": "Legal obligations",
    "Vital Interest": "Vital interests",
    "Vital interests": "Vital interests",
    "Public Interest": "Public interest",
    "Legitimate Interests": "Legitimate interests",
    "Legitimate interests": "Legitimate interests",
}

# Language/casing is changing
special_category_map: Dict = {
    "Consent": "Explicit consent",
    "Employment": "Employment, social security and social protection",
    "Vital Interests": "Vital interests",
    "Non-profit Bodies": "Not-for-profit bodies",
    "Public by Data Subject": "Made public by the data subject",
    "Legal Claims": "Legal claims or judicial acts",
    "Substantial Public Interest": "Reasons of substantial public interest (with a basis in law)",
    "Medical": "Health or social care (with a basis in law)",
    "Public Health Interest": "Public health (with a basis in law)",
}


def _get_privacy_declaration_legal_basis_for_processing(
    data_use_legal_basis: Optional[str],
    legitimate_interest_impact_assessment: Optional[str],
) -> Optional[str]:
    """Use DataUse.legal_basis or DataUse.legitimate_interest_impact_assessment to set the value of
    PrivacyDeclaration.legal_basis_for_processing"""
    if legitimate_interest_impact_assessment:
        return "Legitimate interests"
    if data_use_legal_basis:
        return legal_basis_for_processing_map.get(data_use_legal_basis, None)
    return None


def _get_third_party_data(
    recipients: Optional[List[str]],
) -> Tuple[bool, Optional[str]]:
    """Migrate data from DataUse.recipients field to new PrivacyDeclaration.third_parties and
    PrivacyDeclaration.data_shared_with_third_parties be true
    """
    data_shared_with_third_parties: bool = bool(recipients)
    third_parties: Optional[str] = None

    if isinstance(recipients, list):
        third_parties = "; ".join([elem for elem in recipients if elem]) or None

    return data_shared_with_third_parties, third_parties


def _get_retention_period(
    retention_periods: Optional[List[str]],
) -> str:
    """Migrate data from Dataset.retention field all related PrivacyDeclaration.retention_periods"""
    if not retention_periods:
        return "No retention or erasure policy"

    for retention_period in retention_periods:
        # There may be multiple datasets associated with a privacy declaration.  Choosing the first retention
        # period that does not equal "No retention or erasure policy" if applicable.
        if (
            retention_period is not None
            and retention_period != "No retention or erasure policy"
        ):
            return retention_period
    return "No retention or erasure policy"


def privacy_declaration_additive_migration(bind: Connection):
    """Copy/transform existing data from DataUse and Datasets fields to new Privacy Declaration fields"""
    all_privacy_declarations: Dict[str, Dict] = {}

    # Selecting relevant DataUse fields to move over to PrivacyDeclaration
    get_privacy_declarations_query: TextClause = text(
        """
        SELECT 
            privacydeclaration.id,
            ctl_data_uses.legal_basis,
            ctl_data_uses.legitimate_interest_impact_assessment,
            ctl_data_uses.special_category,
            ctl_data_uses.recipients
        FROM privacydeclaration
        LEFT JOIN ctl_data_uses ON privacydeclaration.data_use = ctl_data_uses.fides_key;
        """
    )
    for row in bind.execute(get_privacy_declarations_query):
        data_shared_with_third_parties, third_parties = _get_third_party_data(
            row["recipients"]
        )

        privacy_declaration_data = {
            LEGAL_BASIS_FOR_PROCESSING: _get_privacy_declaration_legal_basis_for_processing(
                row["legal_basis"], row["legitimate_interest_impact_assessment"]
            ),
            IMPACT_ASSESSMENT_LOCATION: row["legitimate_interest_impact_assessment"],
            PROCESSES_SPECIAL_CATEGORY_DATA: bool(row["special_category"]),
            SPECIAL_CATEGORY_LEGAL_BASIS: special_category_map.get(
                row["special_category"], None
            ),
            DATA_SHARED_WITH_THIRD_PARTIES: data_shared_with_third_parties,
            THIRD_PARTIES: third_parties,
        }

        all_privacy_declarations[row["id"]] = privacy_declaration_data

    # Select first retention periods from associated datasets to migrate over to PrivacyDeclaration
    data_retention_query: TextClause = text(
        """
        SELECT 
            privacydeclaration.id,
            array_agg(ctl_datasets.retention) as retention_periods
        FROM privacydeclaration
        LEFT JOIN ctl_datasets ON ctl_datasets.fides_key = ANY(privacydeclaration.dataset_references)
        GROUP BY privacydeclaration.id;
        """
    )

    for row in bind.execute(data_retention_query):
        all_privacy_declarations[row["id"]][RETENTION_PERIOD] = _get_retention_period(
            row["retention_periods"]
        )

    for privacy_declaration_id, update_data in all_privacy_declarations.items():
        update_privacy_declarations_query: TextClause = text(
            """
                UPDATE privacydeclaration
                SET 
                    legal_basis_for_processing = :legal_basis_for_processing,
                    impact_assessment_location = :impact_assessment_location,
                    processes_special_category_data = :processes_special_category_data,
                    special_category_legal_basis = :special_category_legal_basis,
                    data_shared_with_third_parties = :data_shared_with_third_parties,
                    third_parties = :third_parties,
                    retention_period = :retention_period
                WHERE id = :id;
            """
        )

        bind.execute(
            update_privacy_declarations_query,
            {
                "id": privacy_declaration_id,
                **update_data,
            },
        )


def upgrade():
    bind: Connection = op.get_bind()

    result = bind.execute("SHOW server_version;")  # 12.16 (Debian 12.16-1.pgdg120+1)
    version = result.fetchone()[0]
    major_version = int(version.split(".")[0])

    # Add a new function to be able to combine potentially multiple arrays of different sizes alongside null values
    # into a single array
    if major_version >= 14:
        bind.execute(text("DROP AGGREGATE IF EXISTS array_accum(anycompatiblearray);"))
        bind.execute(
            text(
                """
                CREATE AGGREGATE array_accum (anycompatiblearray)
                (
                    sfunc = array_cat,
                    stype = anycompatiblearray,
                    initcond = '{}'
                );  
                """
            )
        )
    else:
        bind.execute(text("DROP AGGREGATE IF EXISTS array_accum(anyarray);"))
        bind.execute(
            text(
                """
                CREATE AGGREGATE array_accum (anyarray)
                (
                    sfunc = array_cat,
                    stype = anyarray,
                    initcond = '{}'
                );  
                """
            )
        )
    system_dictionary_additive_migration(bind)
    privacy_declaration_additive_migration(bind)


def downgrade():
    pass



from fides.api.schemas.partitioning.bigquery_time_based_partitioning import (
    BigQueryTimeBasedPartitioning,
)
from fides.api.schemas.partitioning.time_based_partitioning import (
    TIME_BASED_REQUIRED_KEYS,
    TimeBasedPartitioning,
    combine_partitions,
    validate_partitioning_list,
)

__all__ = [
    "BigQueryTimeBasedPartitioning",
    "TIME_BASED_REQUIRED_KEYS",
    "TimeBasedPartitioning",
    "combine_partitions",
    "validate_partitioning_list",
]



from fides.common.api.scope_registry import (
    CONNECTION_CREATE_OR_UPDATE,
    CONNECTION_DELETE,
    CONNECTION_READ,
    SYSTEM_DELETE,
    SYSTEM_UPDATE,
)

# System managers are separate from roles, because you are just granted these
# permissions to specific system(s) not to all resources of a given type
SYSTEM_MANAGER_SCOPES = [
    SYSTEM_UPDATE,
    SYSTEM_DELETE,
    CONNECTION_CREATE_OR_UPDATE,
    CONNECTION_READ,
    CONNECTION_DELETE,
]



"""additional tcf fields system

Revision ID: c5a218831820
Revises: 81226042d7d4
Create Date: 2023-10-05 15:40:09.294013

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c5a218831820"
down_revision = "81226042d7d4"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "ctl_systems",
        sa.Column("cookie_max_age_seconds", sa.Integer(), nullable=True),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "uses_cookies",
            sa.Boolean(),
            nullable=False,
            server_default="f",
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "cookie_refresh",
            sa.Boolean(),
            nullable=False,
            server_default="f",
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "uses_non_cookie_access",
            sa.Boolean(),
            nullable=False,
            server_default="f",
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column("legitimate_interest_disclosure_url", sa.String(), nullable=True),
    )

    op.add_column(
        "privacydeclaration",
        sa.Column(
            "flexible_legal_basis_for_processing",
            sa.Boolean(),
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column(
        "ctl_systems",
        "cookie_max_age_seconds",
    )
    op.drop_column(
        "ctl_systems",
        "uses_cookies",
    )
    op.drop_column(
        "ctl_systems",
        "cookie_refresh",
    )
    op.drop_column(
        "ctl_systems",
        "uses_non_cookie_access",
    )
    op.drop_column(
        "ctl_systems",
        "legitimate_interest_disclosure_url",
    )
    op.drop_column(
        "privacydeclaration",
        "flexible_legal_basis_for_processing",
    )



from typing import List

from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig


class PostgresQueryConfig(SQLQueryConfig):
    """
    Generates SQL valid for Postgres
    """

    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """Returns a query string with double quotation mark formatting for tables that have the same names as
        Postgres reserved words."""
        return f'SELECT {field_list} FROM "{self.node.collection.name}" WHERE ({" OR ".join(clauses)})'



"""create ManualTaskConditionalDependency table

Revision ID: d0031087eacb
Revises: 9e3dab26f75f
Create Date: 2025-07-17 21:28:20.937945

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d0031087eacb"
down_revision = "9e3dab26f75f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.create_table(
        "manual_task_conditional_dependency",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("manual_task_id", sa.String(), nullable=False),
        sa.Column("parent_id", sa.String(), nullable=True),
        sa.Column(
            "condition_type",
            sa.String(),  # Use String instead of Enum to avoid auto-creation
            nullable=False,
        ),
        sa.Column("field_address", sa.String(), nullable=True),
        sa.Column("operator", sa.String(), nullable=True),
        sa.Column("value", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column("logical_operator", sa.String(), nullable=True),
        sa.Column("sort_order", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["manual_task_id"], ["manual_task.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["parent_id"], ["manual_task_conditional_dependency.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_index(
        "ix_manual_task_conditional_dependency_condition_type",
        "manual_task_conditional_dependency",
        ["condition_type"],
        unique=False,
    )
    op.create_index(
        "ix_manual_task_conditional_dependency_manual_task_id",
        "manual_task_conditional_dependency",
        ["manual_task_id"],
        unique=False,
    )
    op.create_index(
        "ix_manual_task_conditional_dependency_parent_id",
        "manual_task_conditional_dependency",
        ["parent_id"],
        unique=False,
    )
    op.create_index(
        "ix_manual_task_conditional_dependency_sort_order",
        "manual_task_conditional_dependency",
        ["sort_order"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "ix_manual_task_conditional_dependency_sort_order",
        table_name="manual_task_conditional_dependency",
    )
    op.drop_index(
        "ix_manual_task_conditional_dependency_parent_id",
        table_name="manual_task_conditional_dependency",
    )
    op.drop_index(
        "ix_manual_task_conditional_dependency_manual_task_id",
        table_name="manual_task_conditional_dependency",
    )
    op.drop_index(
        "ix_manual_task_conditional_dependency_condition_type",
        table_name="manual_task_conditional_dependency",
    )

    op.drop_table("manual_task_conditional_dependency")
    # ### end Alembic commands ###



"""
Removes the  stagedresource.child_diff_statuses column.

This migration is run _after_ the data migration
that populates the staged resource ancestor link table,
to prevent unintended data loss.

Revision ID: c586a56c25e7
Revises: bf713b5a021d
Create Date: 2025-06-03 09:47:11.389652

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "c586a56c25e7"
down_revision = "bf713b5a021d"
branch_labels = None
depends_on = None


def upgrade():
    # remove the StagedResource.child_diff_statuses column as it's no longer needed
    op.drop_column("stagedresource", "child_diff_statuses")


def downgrade():
    # re-add the StagedResource.child_diff_statuses column
    op.add_column(
        "stagedresource",
        sa.Column(
            "child_diff_statuses",
            postgresql.JSONB(astext_type=sa.Text()),
            server_default="{}",
            nullable=False,
        ),
    )



from typing import Any, Optional

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import AwaitingAsyncTaskCallback
from fides.api.models.attachment import AttachmentType
from fides.api.models.manual_task import (
    ManualTask,
    ManualTaskConfig,
    ManualTaskConfigurationType,
    ManualTaskEntityType,
    ManualTaskFieldType,
    ManualTaskInstance,
    StatusType,
)
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.task.graph_task import GraphTask, retry
from fides.api.task.manual.manual_task_address import ManualTaskAddress
from fides.api.task.manual.manual_task_utils import (
    get_manual_task_for_connection_config,
)
from fides.api.util.collection_util import Row


class ManualTaskGraphTask(GraphTask):
    """GraphTask implementation for ManualTask execution"""

    @retry(action_type=ActionType.access, default_return=[])
    def access_request(self, *inputs: list[Row]) -> list[Row]:
        """
        Execute manual task logic following the standard GraphTask pattern:
        1. Create ManualTaskInstances if they don't exist
        2. Check for submissions
        3. Return data if submitted, raise AwaitingAsyncTaskCallback if not
        """
        db = self.resources.session
        collection_address = self.execution_node.address

        # Verify this is a manual task address
        if not ManualTaskAddress.is_manual_task_address(collection_address):
            raise ValueError(f"Invalid manual task address: {collection_address}")

        connection_key = ManualTaskAddress.get_connection_key(collection_address)

        # Get manual tasks for this connection
        manual_task = get_manual_task_for_connection_config(db, connection_key)

        if not manual_task:
            return []

        # Check if any manual tasks have ACCESS configs
        # TODO: This will be changed with Manual Task Dependencies Implementation.

        has_access_configs = [
            config
            for config in manual_task.configs
            if config.is_current
            and config.config_type == ManualTaskConfigurationType.access_privacy_request
        ]

        if not has_access_configs:
            # No access configs - complete immediately
            self.log_end(ActionType.access)
            return []

        if not self.resources.request.policy.get_rules_for_action(
            action_type=ActionType.access
        ):
            # TODO: This will be changed with Manual Task Dependencies Implementation.
            self.log_end(ActionType.access)
            return []

        # Check/create manual task instances for ACCESS configs only
        self._ensure_manual_task_instances(
            db,
            manual_task,
            self.resources.request,
            ManualTaskConfigurationType.access_privacy_request,
        )

        # Check if all manual task instances have submissions for ACCESS configs only
        submitted_data = self._get_submitted_data(
            db,
            manual_task,
            self.resources.request,
            ManualTaskConfigurationType.access_privacy_request,
        )

        if submitted_data is not None:
            result: list[Row] = [submitted_data] if submitted_data else []
            self.request_task.access_data = result

            # Mark request task as complete and write execution log
            self.log_end(ActionType.access)
            return result

        # Set privacy request status to requires_input if not already set
        if self.resources.request.status != PrivacyRequestStatus.requires_input:
            self.resources.request.status = PrivacyRequestStatus.requires_input
            self.resources.request.save(db)

        # This should trigger log_awaiting_processing via the @retry decorator
        raise AwaitingAsyncTaskCallback(
            f"Manual task for {connection_key} requires user input"
        )

    def _ensure_manual_task_instances(
        self,
        db: Session,
        manual_task: ManualTask,
        privacy_request: PrivacyRequest,
        allowed_config_type: "ManualTaskConfigurationType",
    ) -> None:
        """Create ManualTaskInstances for configs matching `allowed_config_type` if they don't exist."""

        # ------------------------------------------------------------------
        # Check if instances already exist for this task & entity with the SAME config type
        # This prevents duplicates when configurations are versioned after the privacy
        # request has started, while allowing different config types (access vs erasure)
        # to have separate instances.
        # ------------------------------------------------------------------
        existing_task_instance = (
            db.query(ManualTaskInstance)
            .join(ManualTaskInstance.config)  # Join to access config information
            .filter(
                ManualTaskInstance.task_id == manual_task.id,
                ManualTaskInstance.entity_id == privacy_request.id,
                ManualTaskInstance.entity_type == ManualTaskEntityType.privacy_request,
                # Only check for instances of the same config type
                ManualTaskConfig.config_type == allowed_config_type,
            )
            .first()
        )
        if existing_task_instance:
            # An instance already exists for this privacy request and config type – no need
            # to create another one tied to a newer config version.
            return

        # Check each active config for instances (now we know none exist yet for this config type)
        for config in manual_task.configs:
            if not config.is_current or config.config_type != allowed_config_type:
                # Skip configs that are not current or not relevant for this request type
                continue

            ManualTaskInstance.create(
                db=db,
                data={
                    "task_id": manual_task.id,
                    "config_id": config.id,
                    "entity_id": privacy_request.id,
                    "entity_type": ManualTaskEntityType.privacy_request.value,
                    "status": StatusType.pending.value,
                },
            )

    # pylint: disable=too-many-branches,too-many-nested-blocks
    def _get_submitted_data(
        self,
        db: Session,
        manual_task: ManualTask,
        privacy_request: PrivacyRequest,
        allowed_config_type: "ManualTaskConfigurationType",
    ) -> Optional[dict[str, Any]]:
        """
        Check if all manual task instances have submissions for ALL fields and return aggregated data
        Returns None if any field submissions are missing (all fields must be completed or skipped)
        """
        aggregated_data: dict[str, Any] = {}

        def _format_size(size_bytes: int) -> str:
            units = ["B", "KB", "MB", "GB", "TB"]
            size = float(size_bytes)
            for unit in units:
                if size < 1024.0:
                    return f"{size:.1f} {unit}"
                size /= 1024.0
            return f"{size:.1f} PB"

        candidate_instances: list[ManualTaskInstance] = (
            db.query(ManualTaskInstance)
            .filter(
                ManualTaskInstance.task_id == manual_task.id,
                ManualTaskInstance.entity_id == privacy_request.id,
                ManualTaskInstance.entity_type == ManualTaskEntityType.privacy_request,
            )
            .all()
        )

        if not candidate_instances:
            return None  # No instance yet for this manual task

        for inst in candidate_instances:
            # Skip instances tied to other request types
            if not inst.config or inst.config.config_type != allowed_config_type:
                continue

            all_fields = inst.config.field_definitions or []

            # Every field must have a submission
            if not all(inst.get_submission_for_field(f.id) for f in all_fields):
                return None  # At least one instance still incomplete

            # Ensure status set
            if inst.status != StatusType.completed:
                inst.status = StatusType.completed
                inst.save(db)

            # Aggregate submission data from this instance
            for submission in inst.submissions:
                if not submission.field or not submission.field.field_key:
                    continue

                field_key = submission.field.field_key

                if not isinstance(submission.data, dict):
                    continue

                data_dict: dict[str, Any] = submission.data

                field_type = data_dict.get("field_type")

                if field_type == ManualTaskFieldType.attachment.value:
                    attachment_map: dict[str, dict[str, Any]] = {}
                    for attachment in submission.attachments or []:
                        if (
                            attachment.attachment_type
                            == AttachmentType.include_with_access_package
                        ):
                            try:
                                size, url = attachment.retrieve_attachment()
                                attachment_map[attachment.file_name] = {
                                    "url": str(url) if url else None,
                                    "size": (_format_size(size) if size else "Unknown"),
                                }
                            except (
                                Exception
                            ) as exc:  # pylint: disable=broad-exception-caught
                                logger.warning(
                                    "Error retrieving attachment {}: {}",
                                    attachment.file_name,
                                    str(exc),
                                )

                    aggregated_data[field_key] = attachment_map or None
                else:
                    aggregated_data[field_key] = data_dict.get("value")

        return aggregated_data if aggregated_data else None

    def dry_run_task(self) -> int:
        """Return estimated row count for dry run - manual tasks don't have predictable counts"""
        return 1  # Placeholder - manual tasks generate variable data

    # NEW METHOD: Provide erasure support for manual tasks
    @retry(action_type=ActionType.erasure, default_return=0)
    def erasure_request(
        self,
        retrieved_data: list[Row],
        *erasure_prereqs: int,  # noqa: D401, pylint: disable=unused-argument
        inputs: Optional[list[list[Row]]] = None,
    ) -> int:
        """Execute manual-task-driven erasure logic.

        Mirrors access_request behaviour but returns the number of rows masked (always 0)
        once all required manual task submissions are present. If submissions are
        incomplete the privacy request is paused awaiting user input.
        """
        db = self.resources.session
        collection_address = self.execution_node.address

        # Validate manual task address
        if not ManualTaskAddress.is_manual_task_address(collection_address):
            raise ValueError(f"Invalid manual task address: {collection_address}")

        connection_key = ManualTaskAddress.get_connection_key(collection_address)

        # Fetch relevant manual tasks for this connection
        manual_task = get_manual_task_for_connection_config(db, connection_key)
        if not manual_task:
            # No manual tasks defined – nothing to erase
            self.log_end(ActionType.erasure)
            return 0

        # Check if any manual tasks have ERASURE configs
        has_erasure_configs = [
            config
            for config in manual_task.configs
            if config.is_current
            and config.config_type
            == ManualTaskConfigurationType.erasure_privacy_request
        ]

        if not has_erasure_configs:
            # No erasure configs - complete immediately
            self.log_end(ActionType.erasure)
            return 0

        # Create ManualTaskInstances for ERASURE configs only
        self._ensure_manual_task_instances(
            db,
            manual_task,
            self.resources.request,
            ManualTaskConfigurationType.erasure_privacy_request,
        )

        # Check for full submissions – reuse helper used by access flow, filtering ERASURE configs
        submissions_complete = self._get_submitted_data(
            db,
            manual_task,
            self.resources.request,
            ManualTaskConfigurationType.erasure_privacy_request,
        )

        # If any field submissions are missing, pause processing
        if submissions_complete is None:
            if self.resources.request.status != PrivacyRequestStatus.requires_input:
                self.resources.request.status = PrivacyRequestStatus.requires_input
                self.resources.request.save(db)
            raise AwaitingAsyncTaskCallback(
                f"Manual erasure task for {connection_key} requires user input"
            )

        # Mark rows_masked = 0 (manual tasks do not mask data directly)
        if self.request_task.id:
            # Storing result for DSR 3.0; SQLAlchemy column typing triggers mypy warning
            self.request_task.rows_masked = 0  # type: ignore[assignment]

        # Mark successful completion
        self.log_end(ActionType.erasure)
        return 0



"""
Contains the code that sets up the API.
"""

import os
import re
import sys
from datetime import datetime, timezone
from logging import WARNING
from time import perf_counter
from typing import AsyncGenerator, Callable, Optional
from urllib.parse import unquote

from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import FileResponse, HTMLResponse, JSONResponse
from fideslog.sdk.python.event import AnalyticsEvent
from loguru import logger
from pyinstrument import Profiler
from starlette.background import BackgroundTask
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY
from uvicorn import Config, Server

import fides
from fides.api.app_setup import (
    check_redis,
    create_fides_app,
    log_startup,
    run_database_startup,
)
from fides.api.common_exceptions import MalisciousUrlException
from fides.api.cryptography.identity_salt import get_identity_salt
from fides.api.middleware import handle_audit_log_resource
from fides.api.migrations.hash_migration_job import initiate_bcrypt_migration_task
from fides.api.migrations.post_upgrade_index_creation import (
    initiate_post_upgrade_index_creation,
)
from fides.api.schemas.analytics import Event, ExtraData

# pylint: disable=wildcard-import, unused-wildcard-import
from fides.api.service.privacy_request.email_batch_service import (
    initiate_scheduled_batch_email_send,
)
from fides.api.service.privacy_request.request_service import (
    initiate_interrupted_task_requeue_poll,
    initiate_poll_for_exited_privacy_request_tasks,
    initiate_scheduled_dsr_data_removal,
)

# pylint: disable=wildcard-import, unused-wildcard-import
from fides.api.service.saas_request.override_implementations import *
from fides.api.tasks.scheduled.scheduler import async_scheduler, scheduler
from fides.api.ui import (
    get_admin_index_as_response,
    get_path_to_admin_ui_file,
    get_ui_file_map,
    match_route,
    path_is_in_ui_directory,
)
from fides.api.util.endpoint_utils import API_PREFIX
from fides.api.util.logger import _log_exception
from fides.cli.utils import FIDES_ASCII_ART
from fides.config import CONFIG, check_required_webserver_config_values

IGNORED_AUDIT_LOG_RESOURCE_PATHS = {"/api/v1/login"}
NEXT_JS_CATCH_ALL_SEGMENTS_RE = r"^\[{1,2}\.\.\.\w+\]{1,2}"  # https://nextjs.org/docs/pages/building-your-application/routing/dynamic-routes#catch-all-segments

VERSION = fides.__version__


async def lifespan(wrapped_app: FastAPI) -> AsyncGenerator[None, None]:
    """Run all of the required setup steps for the webserver.

    **NOTE**: The order of operations here _is_ deliberate
    and must be maintained.
    """
    start_time = perf_counter()
    logger.info("Starting server setup...")

    if not CONFIG.dev_mode:
        sys.tracebacklimit = 0

    log_startup()

    await run_database_startup(wrapped_app)

    check_redis()

    if not scheduler.running:
        scheduler.start()
    if not async_scheduler.running:
        async_scheduler.start()

    # generate and/or cache the identity salt
    get_identity_salt()

    initiate_scheduled_batch_email_send()
    initiate_poll_for_exited_privacy_request_tasks()
    initiate_scheduled_dsr_data_removal()
    initiate_interrupted_task_requeue_poll()
    initiate_bcrypt_migration_task()
    initiate_post_upgrade_index_creation()

    logger.debug("Sending startup analytics events...")
    # Avoid circular imports
    from fides.api.analytics import in_docker_container, send_analytics_event

    await send_analytics_event(
        AnalyticsEvent(
            docker=in_docker_container(),
            event=Event.server_start.value,
            event_created_at=datetime.now(tz=timezone.utc),
        )
    )

    # It's just a random bunch of strings when serialized
    if not CONFIG.logging.serialization:
        logger.info(FIDES_ASCII_ART)

    warn_root_user_enabled()

    logger.info("Fides startup complete! v{}", VERSION)
    startup_time = round(perf_counter() - start_time, 3)
    logger.info("Server setup completed in {} seconds", startup_time)
    yield  # All of this happens before the webserver comes up


app = create_fides_app(lifespan=lifespan)  # type: ignore


if CONFIG.dev_mode:

    @app.middleware("http")
    async def profile_request(request: Request, call_next: Callable) -> Response:
        profiling = request.headers.get("profile-request", False)
        if profiling:
            profiler = Profiler(interval=0.001, async_mode="enabled")
            profiler.start()
            await call_next(request)
            profiler.stop()
            logger.debug("Request Profiled!")
            return HTMLResponse(profiler.output_text(timeline=True, show_all=True))

        return await call_next(request)


@app.middleware("http")
async def dispatch_log_request(request: Request, call_next: Callable) -> Response:
    """
    HTTP Middleware that logs analytics events for each call to Fides endpoints.
    :param request: Request to Fides api
    :param call_next: Callable api endpoint
    :return: Response
    """

    # Only log analytics events for requests that are for API endpoints (i.e. /api/...)
    path = request.url.path
    if (not path.startswith(API_PREFIX)) or (path.endswith("/health")):
        return await call_next(request)

    fides_source: Optional[str] = request.headers.get("X-Fides-Source")
    now: datetime = datetime.now(tz=timezone.utc)
    endpoint = f"{request.method}: {request.url}"

    try:
        response = await call_next(request)
        # HTTPExceptions are considered a handled err by default so are not thrown here.
        # Accepted workaround is to inspect status code of response.
        # More context- https://github.com/tiangolo/fastapi/issues/1840
        response.background = BackgroundTask(
            prepare_and_log_request,
            endpoint,
            request.url.hostname,
            response.status_code,
            now,
            fides_source,
            "HTTPException" if response.status_code >= 400 else None,
        )
        return response

    except Exception as e:
        await prepare_and_log_request(
            endpoint, request.url.hostname, 500, now, fides_source, e.__class__.__name__
        )
        _log_exception(e, CONFIG.dev_mode)
        raise


async def prepare_and_log_request(
    endpoint: str,
    hostname: Optional[str],
    status_code: int,
    event_created_at: datetime,
    fides_source: Optional[str],
    error_class: Optional[str],
) -> None:
    """
    Prepares and sends analytics event provided the user is not opted out of analytics.
    """
    # Avoid circular imports
    from fides.api.analytics import (
        accessed_through_local_host,
        in_docker_container,
        send_analytics_event,
    )

    # this check prevents AnalyticsEvent from being called with invalid endpoint during unit tests
    if CONFIG.user.analytics_opt_out:
        return
    await send_analytics_event(
        AnalyticsEvent(
            docker=in_docker_container(),
            event=Event.endpoint_call.value,
            event_created_at=event_created_at,
            local_host=accessed_through_local_host(hostname),
            endpoint=endpoint,
            status_code=status_code,
            error=error_class or None,
            extra_data=(
                {ExtraData.fides_source.value: fides_source} if fides_source else None
            ),
        )
    )


@app.middleware("http")
async def log_request(request: Request, call_next: Callable) -> Response:
    """Log basic information about every request handled by the server."""
    start = datetime.now()

    # If the request fails, we still want to log it
    try:
        response = await call_next(request)
    except:  # pylint: disable=bare-except
        response = Response(status_code=500)

    handler_time = datetime.now() - start

    # Take the total time in seconds and convert it to milliseconds, rounding to 3 decimal places
    total_time = round(handler_time.total_seconds() * 1000, 3)
    logger.bind(
        method=request.method,
        status_code=response.status_code,
        handler_time=f"{total_time}ms",
        path=request.url.path,
        fides_client=request.headers.get("Fides-Client", "unknown"),
    ).info("Request received")
    return response


# Configure the static file paths last since otherwise it will take over all paths
@app.get("/", tags=["Default"])
def read_index() -> Response:
    """
    Return an index.html at the root path
    """

    return get_admin_index_as_response()


def sanitise_url_path(path: str) -> str:
    """Returns a URL path that does not contain any ../ or //"""
    path = unquote(path)
    path = os.path.normpath(path)

    for token in path.split("/"):
        if ".." in token and not re.search(NEXT_JS_CATCH_ALL_SEGMENTS_RE, token):
            logger.warning(
                f"Potentially dangerous use of URL hierarchy in path: {path}"
            )
            raise MalisciousUrlException()
    return path


@app.get("/{catchall:path}", response_class=Response, tags=["Default"])
def read_other_paths(request: Request) -> Response:
    """
    Return related frontend files. Adapted from https://github.com/tiangolo/fastapi/issues/130
    """
    # check first if requested file exists (for frontend assets)
    path = request.path_params["catchall"]
    logger.debug(f"Catch all path detected: {path}")
    try:
        path = sanitise_url_path(path)
    except MalisciousUrlException:
        # if a maliscious URL is detected, route the user to the index
        return get_admin_index_as_response()

    # search for matching route in package (i.e. /dataset)
    ui_file = match_route(get_ui_file_map(), path)

    # if not, check if the requested file is an asset (i.e. /_next/static/...)
    if not ui_file:
        ui_file = get_path_to_admin_ui_file(path)

    # Serve up the file as long as it is within the UI directory
    if ui_file and ui_file.is_file():
        if not path_is_in_ui_directory(ui_file):
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Item not found"
            )
        logger.debug(
            "catchall request path '{}' matched static admin UI file: {}",
            path,
            ui_file,
        )
        return FileResponse(ui_file)

    # raise 404 for anything that should be backend endpoint but we can't find it
    if path.startswith(API_PREFIX[1:]):
        logger.debug(
            "catchall request path '{}' matched an invalid API route, return 404",
            path,
        )
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Item not found"
        )

    # otherwise return the index
    logger.debug(
        "catchall request path '{}' did not match any admin UI routes, return generic admin UI index",
        path,
    )
    return get_admin_index_as_response()


def warn_root_user_enabled() -> None:
    """
    Log a startup warning if root user is enabled.
    Extracted as a function because this may need to be done in multiple places,
    depending on how the server is started.
    """
    if CONFIG.security.root_username or CONFIG.security.oauth_root_client_id:
        logger.warning(
            "Root Username & Password are configured and can be used to login as a root user. If unexpected, review security settings (FIDES__SECURITY__ROOT_USERNAME and FIDES__SECURITY__ROOT_PASSWORD)"
        )


def start_webserver(port: int = 8080) -> None:
    """Run the webserver."""
    check_required_webserver_config_values(config=CONFIG)
    server = Server(Config(app, host="0.0.0.0", port=port, log_level=WARNING))

    logger.info(
        "Starting webserver - Host: {}, Port: {}, Log Level: {}",
        server.config.host,
        server.config.port,
        server.config.log_level,
    )

    warn_root_user_enabled()

    server.run()


@app.middleware("http")
async def action_to_audit_log(
    request: Request,
    call_next: Callable,
) -> Response:
    """Log basic information about every non-GET request handled by the server."""

    if (
        request.method != "GET"
        and request.scope["path"] not in IGNORED_AUDIT_LOG_RESOURCE_PATHS
        and CONFIG.security.enable_audit_log_resource_middleware
    ):
        try:
            await handle_audit_log_resource(request)
        except Exception as exc:
            logger.debug(exc)
    return await call_next(request)


@app.exception_handler(RequestValidationError)
async def request_validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    """Prevents field input and pydantic error message URL from being returned in the response
    For example, if someone is creating a user and the request fails, this prevents the user's
    password from being returned in the error message
    """
    return JSONResponse(
        status_code=HTTP_422_UNPROCESSABLE_ENTITY,
        content={
            "detail": jsonable_encoder(exc.errors(), exclude={"input", "url", "ctx"})
        },
    )



import uuid
from enum import Enum

from loguru import logger
from sqlalchemy.orm import Query, Session

from fides.api.common_exceptions import MessageDispatchException
from fides.api.models.policy import Policy, Rule
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType, CurrentStep
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.service.connectors import get_connector
from fides.api.service.privacy_request.request_runner_service import (
    get_consent_email_connection_configs,
    get_erasure_email_connection_configs,
)
from fides.api.tasks import DatabaseTask, celery_app
from fides.api.tasks.scheduled.scheduler import create_cron_trigger, scheduler
from fides.api.util.lock import redis_lock
from fides.config import get_config
from fides.service.privacy_request.privacy_request_service import queue_privacy_request

CONFIG = get_config()
BATCH_EMAIL_SEND = "batch_email_send"
BATCH_EMAIL_SEND_LOCK = "batch_email_send_lock"
BATCH_EMAIL_SEND_LOCK_TIMEOUT = 600


class EmailExitState(Enum):
    """A schema to describe where the email send process exited.
    For logging and testing"""

    no_applicable_privacy_requests = "no_applicable_privacy_requests"
    no_applicable_connectors = "no_applicable_connectors"
    missing_required_data = "missing_required_data"
    email_send_failed = "email_send_failed"
    complete = "complete"
    email_send_already_running = "email_send_already_running"


@celery_app.task(base=DatabaseTask, bind=True)
def send_email_batch(self: DatabaseTask) -> EmailExitState:
    """Sends emails for each relevant connector with applicable user details batched together."""
    with redis_lock(
        lock_key=BATCH_EMAIL_SEND_LOCK, timeout=BATCH_EMAIL_SEND_LOCK_TIMEOUT
    ) as lock:
        if not lock:
            return EmailExitState.email_send_already_running

        batch_id = str(uuid.uuid4())
        logger.info("Starting batch email send {}...", batch_id)
        with self.get_new_session() as session:
            privacy_requests: Query = (
                session.query(PrivacyRequest)
                .filter(
                    PrivacyRequest.status == PrivacyRequestStatus.awaiting_email_send
                )
                .filter(PrivacyRequest.deleted_at.is_(None))
                .order_by(PrivacyRequest.created_at.asc())  # oldest first
            )
            if not privacy_requests.first():
                logger.info(
                    "Skipping batch email send with status: {}",
                    EmailExitState.no_applicable_privacy_requests.value,
                )
                return EmailExitState.no_applicable_privacy_requests

            consent_configs: Query = get_consent_email_connection_configs(session)
            erasure_configs: Query = get_erasure_email_connection_configs(session)
            combined_configs = consent_configs.union_all(erasure_configs)
            if not combined_configs.first():
                requeue_privacy_requests_after_email_send(privacy_requests, session)
                logger.info(
                    "Skipping batch email send with status: {}",
                    EmailExitState.no_applicable_connectors.value,
                )
                return EmailExitState.no_applicable_connectors

            try:
                # erasure
                for connection_config in erasure_configs:
                    get_connector(connection_config).batch_email_send(  # type: ignore
                        filter_privacy_requests_by_action_type(
                            privacy_requests, ActionType.erasure
                        ),
                        batch_id,
                    )

                # consent
                for connection_config in consent_configs:
                    get_connector(connection_config).batch_email_send(  # type: ignore
                        filter_privacy_requests_by_action_type(
                            privacy_requests, ActionType.consent
                        ),
                        batch_id,
                    )
            except MessageDispatchException as exc:
                logger.error(
                    "Batch email send for connector failed with exception: '{}'",
                    exc,
                )
                return EmailExitState.email_send_failed

        requeue_privacy_requests_after_email_send(privacy_requests, session)
        return EmailExitState.complete


def filter_privacy_requests_by_action_type(
    privacy_requests: Query, action_type: ActionType
) -> Query:
    """Applies a filter for the specific Rule.action_type to the passed in privacy_requests."""
    return (
        privacy_requests.join(Policy, PrivacyRequest.policy_id == Policy.id)
        .join(Rule, Policy.id == Rule.policy_id)
        .filter(Rule.action_type == action_type)
    )


def requeue_privacy_requests_after_email_send(
    privacy_requests: Query, db: Session
) -> None:
    """After batch consent email send, requeue privacy requests from the post webhooks step
    to wrap up processing and transition to a "complete" state.

    Also cache on the privacy request itself that it is paused at the post-webhooks state,
    in case something happens in re-queueing.
    """
    logger.info("Batched email send complete.")
    logger.info("Queuing privacy requests from 'post_webhooks' step.")
    for privacy_request in privacy_requests:
        privacy_request.cache_paused_collection_details(
            step=CurrentStep.post_webhooks,
            collection=None,
            action_needed=None,
        )
        privacy_request.status = PrivacyRequestStatus.paused
        privacy_request.save(db=db)

        queue_privacy_request(
            privacy_request_id=privacy_request.id,
            from_step=CurrentStep.post_webhooks.value,
        )


def initiate_scheduled_batch_email_send() -> None:
    """Initiates scheduler to add weekly batch email send"""

    if CONFIG.test_mode:
        return

    assert scheduler.running, "Scheduler is not running! Cannot add Batch Email job."

    cron_trigger = create_cron_trigger(
        cron_expression=CONFIG.execution.email_send_cron_expression,
        timezone=CONFIG.execution.email_send_timezone,
    )

    logger.info("Initiating scheduler for batch email send")
    scheduler.add_job(
        func=send_email_batch,
        kwargs={},
        id=BATCH_EMAIL_SEND,
        coalesce=False,
        replace_existing=True,
        trigger=cron_trigger,
    )



import zipfile
from io import BytesIO
from typing import Any, Optional

import pandas as pd

from fides.api.tasks.encryption_utils import encrypt_access_request_results
from fides.config import CONFIG


def create_csv_from_dataframe(df: pd.DataFrame) -> BytesIO:
    """Create a CSV file from a pandas DataFrame.

    Args:
        df: The DataFrame to convert to CSV

    Returns:
        BytesIO: A file-like object containing the CSV data
    """
    buffer = BytesIO()
    df.to_csv(buffer, index=False, encoding=CONFIG.security.encoding)
    buffer.seek(0)
    return buffer


def create_attachment_csv(attachments: list[dict[str, Any]]) -> Optional[BytesIO]:
    """Create a CSV file containing attachment metadata.

    Args:
        attachments: List of attachment dictionaries
        privacy_request_id: The ID of the privacy request for encryption

    Returns:
        Optional[BytesIO]: A file-like object containing the CSV data, or None if no attachments
    """
    if not attachments:
        return None

    # Filter out invalid attachments and create a list of valid ones
    valid_attachments = []
    for a in attachments:
        if not isinstance(a, dict):
            continue

        # Check if the attachment has at least one of the required fields
        if not any(
            key in a
            for key in ["file_name", "file_size", "content_type", "download_url"]
        ):
            continue

        valid_attachments.append(
            {
                "file_name": a.get("file_name", ""),
                "file_size": a.get("file_size", 0),
                "content_type": a.get("content_type", "application/octet-stream"),
                "download_url": a.get("download_url", ""),
            }
        )

    # Return None if there are no valid attachments
    if not valid_attachments:
        return None

    df = pd.DataFrame(valid_attachments)

    if df.empty:
        return None

    return create_csv_from_dataframe(df)


def _write_attachment_csv(
    zip_file: zipfile.ZipFile,
    key: str,
    idx: int,
    attachments: list[dict[str, Any]],
    privacy_request_id: str,
) -> None:
    """Write attachment data to a CSV file in the zip archive.

    Args:
        zip_file: The zip file to write to
        key: The key for the data
        idx: The index of the item in the list
        attachments: List of attachment dictionaries
        privacy_request_id: The ID of the privacy request for encryption
    """
    buffer = create_attachment_csv(attachments)
    if buffer:
        zip_file.writestr(
            f"{key}/{idx + 1}/attachments.csv",
            encrypt_access_request_results(buffer.getvalue(), privacy_request_id),
        )


def _write_item_csv(
    zip_file: zipfile.ZipFile,
    key: str,
    items: list[dict[str, Any]],
    privacy_request_id: str,
) -> None:
    """Write item data to a CSV file in the zip archive.

    Args:
        zip_file: The zip file to write to
        key: The key for the data
        items: List of items to write
        privacy_request_id: The ID of the privacy request for encryption
    """
    if items:
        df = pd.DataFrame(items)
        buffer = create_csv_from_dataframe(df)
        zip_file.writestr(
            f"{key}.csv",
            encrypt_access_request_results(buffer.getvalue(), privacy_request_id),
        )


def _write_simple_csv(
    zip_file: zipfile.ZipFile,
    key: str,
    value: Any,
    privacy_request_id: str,
) -> None:
    """Write simple key-value data to a CSV file in the zip archive.

    Args:
        zip_file: The zip file to write to
        key: The key for the data
        value: The value to write
        privacy_request_id: The ID of the privacy request for encryption
    """
    df = pd.json_normalize({key: value})
    buffer = create_csv_from_dataframe(df)
    zip_file.writestr(
        f"{key}.csv",
        encrypt_access_request_results(buffer.getvalue(), privacy_request_id),
    )


def write_csv_to_zip(
    zip_file: zipfile.ZipFile, data: dict[str, Any], privacy_request_id: str
) -> None:
    """Write data to a zip file in CSV format.

    Args:
        zip_file: The zip file to write to
        data: The data to convert to CSV
        privacy_request_id: The ID of the privacy request for encryption
    """
    for key, value in data.items():
        if (
            isinstance(value, list)
            and value
            and all(isinstance(item, dict) for item in value)
        ):
            # Handle lists of dictionaries
            items: list[dict[str, Any]] = []
            for item in value:
                # Extract attachments if they exist
                attachments = item.pop("attachments", [])
                if attachments:
                    _write_attachment_csv(
                        zip_file, key, len(items), attachments, privacy_request_id
                    )
                items.append(item)
            _write_item_csv(zip_file, key, items, privacy_request_id)
        else:
            _write_simple_csv(zip_file, key, value, privacy_request_id)



"""add description column to asset

Revision ID: 9a1eacd2666b
Revises: 69ad6d844e21
Create Date: 2025-02-28 13:59:26.022822

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "9a1eacd2666b"
down_revision = "69ad6d844e21"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("asset", sa.Column("description", sa.String(), nullable=True))


def downgrade():
    op.drop_column("asset", "description")



"""
Manual Task Connector - A minimal connector for manual task operations.

Since manual tasks don't actually connect to external systems, this connector
provides no-op implementations of the BaseConnector interface.
"""

from typing import Any, Dict, List, Optional

from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row


class ManualTaskQueryConfig(QueryConfig):
    """Minimal query config for manual tasks - not actually used"""

    def generate_query(
        self, input_data: Dict[str, List[Any]], policy: Optional[Policy]
    ) -> str:
        return "Manual task: no query needed"

    def dry_run_query(self) -> str:
        return "Manual task: no query needed"

    def query_to_str(self, t: Any, input_data: Dict[str, List[Any]]) -> str:
        """Convert query to string - not used for manual tasks"""
        return "Manual task: no query needed"

    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Any:
        """Generate update statement - not used for manual tasks"""
        return None


class ManualTaskConnector(BaseConnector):
    """
    Minimal connector for manual tasks.

    This connector provides no-op implementations since manual tasks don't
    actually connect to external systems. The actual manual task logic
    is handled by ManualTaskGraphTask.access_request()
    """

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """Return a minimal query config - not actually used for manual tasks"""
        return ManualTaskQueryConfig(node)

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """Manual tasks don't have connections to test"""
        return ConnectionTestStatus.succeeded

    def create_client(self) -> None:
        """Manual tasks don't need database clients"""
        return None

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """
        This method is not used for manual tasks.
        Manual task data retrieval is handled by ManualTaskGraphTask.access_request()
        """
        return []

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """
        Manual tasks don't support erasure operations.
        Manual tasks are for data collection, not data modification.
        """
        return 0

    def close(self) -> None:
        """No resources to close for manual tasks"""

    @property
    def requires_primary_keys(self) -> bool:
        """Manual tasks don't require primary keys since they don't modify data"""
        return False



"""add user login and password reset dates

Revision ID: c9759675b5d3
Revises: 906d7198df28
Create Date: 2022-04-14 13:16:58.940571

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c9759675b5d3"
down_revision = "906d7198df28"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "fidesopsuser",
        sa.Column("last_login_at", sa.DateTime(), nullable=True),
    )
    op.add_column(
        "fidesopsuser", sa.Column("password_reset_at", sa.DateTime(), nullable=True)
    )


def downgrade():
    op.drop_column("fidesopsuser", "last_login_at")
    op.drop_column("fidesopsuser", "password_reset_at")



from enum import Enum


class SystemType(Enum):
    data_catalog = "data_catalog"
    database = "database"
    email = "email"
    manual = "manual"
    saas = "saas"
    system = "system"
    website = "website"
